"""Access_key types file."""

import datetime
from pyramid.security import (
    Allow,
    Deny,
    Authenticated,
    Everyone,
)
from pyramid.settings import asbool
from pyramid.view import view_config
from .. import (
    collection,
    load_schema,
)
from ..authentication import (
    generate_password,
    generate_user,
    CRYPT_CONTEXT,
)
from ..crud_views import (
    collection_add,
    item_edit,
)
from ..project_app import app_project
from ..validators import (
    validate_item_content_post,
)
from ..util import debug_log
from .base import (
    Item,
    DELETED_ACL,
    ONLY_ADMIN_VIEW_ACL,
)


@collection(
    name='access-keys',
    unique_key='access_key:access_key_id',
    properties={
        'title': 'Access keys',
        'description': 'Programmatic access keys',
    },
    acl=[
        (Allow, Authenticated, 'add'),
        (Allow, 'group.admin', 'list'),
        (Allow, 'group.read-only-admin', 'list'),
        (Allow, 'remoteuser.INDEXER', 'list'),
        (Allow, 'remoteuser.EMBED', 'list'),
        (Deny, Everyone, 'list'),
    ])
class AccessKey(Item):
    """AccessKey class."""
    ACCESS_KEY_EXPIRATION_TIME = 90  # days
    item_type = 'access_key'
    schema = load_schema('snovault:schemas/access_key.json')
    embedded_list = []

    STATUS_ACL = {
        'current': [(Allow, 'role.owner', ['view', 'edit'])] + ONLY_ADMIN_VIEW_ACL,
        'deleted': DELETED_ACL,
    }

    @classmethod
    def create(cls, registry, uuid, properties, sheets=None):
        """ Sets the access key timeout 90 days from creation. """
        if app_project().access_key_has_expiration_date():
            properties['expiration_date'] = (datetime.datetime.utcnow() + datetime.timedelta(
                days=cls.ACCESS_KEY_EXPIRATION_TIME)).isoformat()
        return super().create(registry, uuid, properties, sheets)

    def __ac_local_roles__(self):
        """grab and return user as owner."""
        owner = 'userid.%s' % self.properties['user']
        return {owner: 'role.owner'}

    def __json__(self, request):
        """delete the secret access key has from the object when used."""
        properties = super(AccessKey, self).__json__(request)
        del properties['secret_access_key_hash']
        return properties

    def update(self, properties, sheets=None):
        """smth."""
        # make sure PUTs preserve the secret access key hash
        if 'secret_access_key_hash' not in properties:
            new_properties = self.properties.copy()
            new_properties.update(properties)
            properties = new_properties
        # set new expiration if applicable
        if app_project().access_key_has_expiration_date():
            properties['expiration_date'] = (datetime.datetime.utcnow() + datetime.timedelta(
                days=self.ACCESS_KEY_EXPIRATION_TIME)).isoformat()
        self._update(properties, sheets)

    class Collection(Item.Collection):
        pass


# access keys have view permissions for update so readonly admin and the like
# can create access keys to download files.
@view_config(context=AccessKey.Collection, request_method='POST',
             permission='add',
             validators=[validate_item_content_post])
@debug_log
def access_key_add(context, request):
    """smth."""
    crypt_context = request.registry[CRYPT_CONTEXT]

    if 'access_key_id' not in request.validated:
        request.validated['access_key_id'] = generate_user()

    if 'user' not in request.validated:
        # request.validated['user'], = [
        #    principal.split('.', 1)[1]
        #    for principal in request.effective_principals
        #    if principal.startswith('userid.')
        # ]
        user = [
            principal.split('.', 1)[1]
            for principal in request.effective_principals
            if principal.startswith('userid.')
        ]
        # Only (at most) one single user expected from above.
        request.validated['user'] = user[0] if len(user) == 1 else None

    password = None
    if 'secret_access_key_hash' not in request.validated:
        password = generate_password()
        request.validated['secret_access_key_hash'] = crypt_context.hash(password)

    result = collection_add(context, request)

    if password is None:
        result['secret_access_key'] = None
    else:
        result['secret_access_key'] = password

    result['access_key_id'] = request.validated['access_key_id']
    result['description'] = request.validated.get('description', "")
    return result


@view_config(name='reset-secret', context=AccessKey,
             permission='add',
             request_method='POST', subpath_segments=0)
@debug_log
def access_key_reset_secret(context, request):
    """smth."""
    request.validated = context.properties.copy()
    crypt_context = request.registry[CRYPT_CONTEXT]
    password = generate_password()
    new_hash = crypt_context.hash(password)
    request.validated['secret_access_key_hash'] = new_hash
    result = item_edit(context, request, render=False)
    result['access_key_id'] = request.validated['access_key_id']
    result['secret_access_key'] = password
    return result


@view_config(context=AccessKey, permission='view_raw', request_method='GET',
             name='raw')
@debug_log
def access_key_view_raw(context, request):
    """smth."""
    if asbool(request.params.get('upgrade', True)):
        properties = context.upgrade_properties()
    else:
        properties = context.properties.copy()
    del properties['secret_access_key_hash']
    return properties



# Factored out of server_defaults to avoid circular dependencies now that get_userid is
# used in resources.Item.is_update_by_admin_user (commonized from fourfront/cgap-portal),
# since server_defaults imports schema_utils which imports resources which wants get_userid.

from snovault.schema_validation import NO_DEFAULT
from pyramid.threadlocal import get_current_request
from .interfaces import COLLECTIONS


def _userid():
    request = get_current_request()
    for principal in request.effective_principals:
        if principal.startswith('userid.'):
            return principal[7:]
    return NO_DEFAULT


def get_userid():
    """ Wrapper for the server_default 'userid' above so it is not called through SERVER_DEFAULTS in our code """
    return _userid()


def get_user_resource():
    request = get_current_request()
    userid_found = _userid()
    if userid_found == NO_DEFAULT:
        return NO_DEFAULT
    return request.registry[COLLECTIONS]['user'][userid_found]



#from ..project_defs import SnovaultProject
from ..project_app import app_project
import argparse
import atexit
import botocore.exceptions
import cgi
import datetime
import elasticsearch
import io
import json
import os
import psycopg2
import re
import signal
import structlog
import threading
import time
import webtest

from dcicutils.env_utils import is_stg_or_prd_env
from dcicutils.misc_utils import VirtualApp, ignored, check_true, full_class_name, environ_bool, PRINT
from pyramid import paster
# Possibly still needed by some commented-out code.
# from pyramid.response import Response
from pyramid.view import view_config
from snovault.util import debug_log
from ..embed import subrequest_object
from ..types.ingestion import SubmissionFolio, IngestionSubmission
from ..util import (
    debuglog, get_trusted_email, beanstalk_env_from_request,
    register_path_content_type, vapp_for_email,  # vapp_for_ingestion,
    SettingsKey, make_s3_client, extra_kwargs_for_s3_encrypt_key_id,
)
from .common import metadata_bundles_bucket, get_parameter, IngestionReport
from .exceptions import UnspecifiedFormParameter, SubmissionFailure  # , BadParameter
from .ingestion_listener_base import (
    DEBUG_SUBMISSIONS,
    IngestionListenerBase,
)
from .ingestion_message_handler_decorator import call_ingestion_message_handler
from .ingestion_processor_decorator import get_ingestion_processor
from .queue_utils import IngestionQueueManager


log = structlog.getLogger(__name__)
EPILOG = __doc__
INGESTION_QUEUE = 'ingestion_queue'


def includeme(config):
    # config.add_route('process_ingestion', '/process_ingestion')
    config.add_route('queue_ingestion', '/queue_ingestion')
    config.add_route('ingestion_status', '/ingestion_status')
    config.add_route('submit_for_ingestion', '/submit_for_ingestion')
    config.registry[INGESTION_QUEUE] = IngestionQueueManager(config.registry)
    config.scan(__name__)


SUBMISSION_PATTERN = re.compile(r'^/ingestion-submissions/([0-9a-fA-F-]+)(|/.*)$')

register_path_content_type(path='/submit_for_ingestion', content_type='multipart/form-data')


def extract_submission_info(request):
    matched = SUBMISSION_PATTERN.match(request.path_info)
    if matched:
        submission_id = matched.group(1)
    else:
        raise SubmissionFailure("request.path_info is not in the expected form: %s" % request.path_info)

    instance = subrequest_object(request, submission_id)
    return submission_id, instance


@view_config(name='submit_for_ingestion', request_method='POST', context=IngestionSubmission,
             # Apparently adding this 'accept' causes discrimination on incoming requests not to find this method.
             # We do want this type, and instead we check the request to make sure we got it, but we omit it here
             # for practical reasons. -kmp 10-Sep-2020
             # accept='multipart/form-data',
             permission='edit')
@debug_log
def submit_for_ingestion(context, request):
    ignored(context)

    check_true(request.content_type == 'multipart/form-data',  # even though we can't declare we accept this
               "Expected request to have content_type 'multipart/form-data'.", error_class=SubmissionFailure)

    bs_env = beanstalk_env_from_request(request)
    bundles_bucket = metadata_bundles_bucket(request.registry)
    datafile = request.POST.get('datafile')
    if datafile is None:
        # S3 protocol; not uploading from here (SubmitCGAP uploads directly).
        # Added circa March 2023 for Fourfront ontology ingestion.
        filename = request.POST['datafile_source_filename']
        override_name = None
    elif isinstance(datafile, cgi.FieldStorage):
        filename = datafile.filename
        override_name = request.POST.get('override_name', None)
    else:
        # e.g., specifically it might be b'' when no file is selected,
        # but IMPORTANTLY, cgi.FieldStorage has no predefined boolean value,
        # so we can't just ask to check 'not datafile'. Sigh. -kmp 5-Aug-2020
        raise UnspecifiedFormParameter('datafile')
    parameters = dict(request.POST)  # Convert to regular dictionary, which is also a copy
    parameters['datafile'] = filename

    # Other parameters, like validate_only, will ride in on parameters via the manifest on s3

    submission_id, instance = extract_submission_info(request)

    # The three arguments award, lab, and ingestion_type were needed in the old protocol
    # but are not needed in the new protocol because someone will have set up the IngestionSubmission
    # object already with the right values. We tolerate them here, but we insist they be consistent (redundant).
    # Note, too, that we use the 'update=True' option that causes them to be added to our parameters if they are
    # missing, defaulted from the previous item, so that they will be written to the parameter block stored on S3.
    # (We could do that differently now, by looking them up dynamically, but rather than risk making a mistake,
    # I just went with path of least resistance for now.)
    # -kmp 2-Dec-2020
    #
    # Same goes for award and lab which were brought from fourfront
    # into this common snovault version of ingestion_listener.py.
    # -dmichaels 12-May-2023

    if instance.get("institution"):
        institution = instance['institution']['@id']
        institution_arg = get_parameter(parameters, "institution", default=institution, update=True)
        if institution_arg != institution:
            # If the "institution" argument was passed, which we no longer require, make sure it's consistent.
            raise SubmissionFailure("'institution' was supplied inconsistently for submit_for_ingestion.")

    if instance.get("project"):
        project = instance['project']['@id']
        project_arg = get_parameter(parameters, "project", default=project, update=True)
        if project_arg != project:
            # If the "project" argument was passed, which we no longer require, make sure it's consistent.
            raise SubmissionFailure("'project' was supplied inconsistently for submit_for_ingestion.")

    if instance.get("award"):
        award = instance['award']['@id']
        award_arg = get_parameter(parameters, "award", default=award, update=True)
        if award_arg != award:
            # If the "award" argument was passed, which we no longer require, make sure it's consistent.
            raise SubmissionFailure("'award' was supplied inconsistently for submit_for_ingestion.")

    if instance.get("lab"):
        lab = instance['lab']['@id']
        lab_arg = get_parameter(parameters, "lab", default=lab, update=True)
        if lab_arg != lab:
            # If the "lab" argument was passed, which we no longer require, make sure it's consistent.
            raise SubmissionFailure("'lab' was supplied inconsistently for submit_for_ingestion.")

    if instance.get("consortia"):
        consortia = instance["consortia"]
        if isinstance(consortia, list) and len(consortia) > 0:
            consortium = consortia[0]
        else:
            consortium = consortia
        consortium = consortium["@id"]
        consortium_arg = get_parameter(parameters, "consortium", default=consortium, update=True)
        if consortium_arg != consortium:
            # If the "consortium" argument was passed, which we no longer require, make sure it's consistent.
            raise SubmissionFailure("'consortium' was supplied inconsistently for submit_for_ingestion.")

    if instance.get("submission_centers"):
        submission_centers = instance["submission_centers"]
        if isinstance(submission_centers, list) and len(submission_centers) > 0:
            submission_center = submission_centers[0]
        else:
            submission_center = submission_centers
        submission_center = submission_center['@id']
        submission_center_arg = get_parameter(parameters, "submission_center", default=submission_center, update=True)
        if submission_center_arg != submission_center:
            # If the "submission_center" argument was passed, which we no longer require, make sure it's consistent.
            raise SubmissionFailure("'submission_center' was supplied inconsistently for submit_for_ingestion.")

    ingestion_type = instance['ingestion_type']
    ingestion_type_arg = get_parameter(parameters, "ingestion_type", default=ingestion_type, update=True)
    if ingestion_type_arg != ingestion_type:
        # If the "ingestion_type" argument was passed, which we no longer require, make sure it's consistent.
        raise SubmissionFailure("'ingestion_type' was supplied inconsistently for submit_for_ingestion.")

    # ``input_file`` contains the actual file data which needs to be
    # stored somewhere.

    if datafile is not None:
        input_file_stream = datafile.file
        input_file_stream.seek(0)

    # NOTE: Some reference information about uploading files to s3 is here:
    #   https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-uploading-files.html

    # submission.set_item_detail(object_name=manifest['object_name'], parameters=manifest['parameters'],
    #                            institution=institution, project=project)

    # submission_id = str(uuid.uuid4())
    if filename.endswith(".gz"):
        # Maintain the real file suffix is .gz file.
        _, ext = os.path.splitext(filename[:-3])
        gz = ".gz"
    else:
        _, ext = os.path.splitext(filename)
        gz = ""
    object_name = "{id}/datafile{ext}{gz}".format(id=submission_id, ext=ext, gz=gz) if datafile is not None else submission_id
    manifest_name = "{id}/manifest.json".format(id=submission_id)

    # We might need to extract some additional information from the GAC
    s3_client = make_s3_client()

    upload_time = datetime.datetime.utcnow().isoformat()
    success = True
    message = "Uploaded successfully."

    # Set up potentially useful additional args
    s3_encrypt_key_id = request.registry.settings.get(SettingsKey.S3_ENCRYPT_KEY_ID)
    extra_kwargs = extra_kwargs_for_s3_encrypt_key_id(s3_encrypt_key_id=s3_encrypt_key_id,
                                                      client_name='submit_for_ingestion')

    if extra_kwargs:
        additional_info = f" (with SSEKMSKeyId: {s3_encrypt_key_id})"
    else:
        additional_info = " (no SSEKMSKeyId)"

    if datafile is not None:
        try:
            # Make sure to pass any extra args.
            s3_client.upload_fileobj(input_file_stream, Bucket=bundles_bucket, Key=object_name, **extra_kwargs)
        except botocore.exceptions.ClientError as e:
            log.error(e)
            success = False
            message = f"{full_class_name(e)}: {str(e)}{additional_info}"

    # This manifest will be stored in the manifest.json file on on s3 AND will be returned from this endpoint call.
    manifest_content = {
        "filename": filename,
        "object_name": object_name,
        "s3_encrypt_key_id": s3_encrypt_key_id,
        "submission_id": submission_id,
        "submission_uri": SubmissionFolio.make_submission_uri(submission_id),
        "beanstalk_env_is_prd": is_stg_or_prd_env(bs_env),
        "beanstalk_env": bs_env,
        "bucket": bundles_bucket,
        "authenticated_userid": request.authenticated_userid,
        "email": get_trusted_email(request, context="Submission", raise_errors=False),
        "success": success,
        "message": message,
        "upload_time": upload_time,
        "parameters": parameters,
    }

    manifest_content_formatted = json.dumps(manifest_content, indent=2)

    if success:

        try:
            with io.BytesIO(manifest_content_formatted.encode('utf-8')) as fp:
                s3_client.upload_fileobj(fp, Bucket=bundles_bucket, Key=manifest_name, **extra_kwargs)
        except botocore.exceptions.ClientError as e:
            log.error(e)
            message = f"{full_class_name(e)} (while uploading metadata): {str(e)}{additional_info}"
            raise SubmissionFailure(message)

        queue_manager = get_queue_manager(request, override_name=override_name)
        _, failed = queue_manager.add_uuids([submission_id], ingestion_type=ingestion_type)

        if failed:
            # If there's a failure, failed will be a list of one problem description since we only submitted one thing.
            raise SubmissionFailure(failed[0])

    if not success:

        raise SubmissionFailure(message)

    return manifest_content


@view_config(route_name='ingestion_status', request_method='GET', permission='index')
@debug_log
def ingestion_status(context, request):
    """ Status route, essentially identical to indexing_status. """
    ignored(context)
    queue_manager = request.registry[INGESTION_QUEUE]
    n_waiting, n_inflight = queue_manager.get_counts()
    return {
        'title': 'Ingestion Status',
        'waiting': n_waiting,
        'inflight': n_inflight
    }



def process_submission(*, submission_id, ingestion_type, app, bundles_bucket=None, s3_client=None):
    ignored(s3_client)  # we might want to restore the ability to pass this, but no one actually does. -kmp 6-Dec-2021
    bundles_bucket = bundles_bucket or metadata_bundles_bucket(app.registry)
    s3_client = make_s3_client()
    manifest_name = "{id}/manifest.json".format(id=submission_id)
    log.warning(f'Processing submission {manifest_name}')
    obj = s3_client.get_object(Bucket=bundles_bucket, Key=manifest_name)
    # data = json.load(obj)['Body']
    data = json.load(obj['Body'])
    email = None
    try:
        email = data['email']
    except KeyError as e:
        ignored(e)
        debuglog("Manifest data is missing 'email' field.")
        if DEBUG_SUBMISSIONS:
            pass
    debuglog("processing submission %s with email %s" % (submission_id, email))
    with vapp_for_email(email=email, app=app) as vapp:
        if DEBUG_SUBMISSIONS:
            PRINT("PROCESSING FOR %s" % email)
        submission = SubmissionFolio(vapp=vapp, ingestion_type=ingestion_type, submission_id=submission_id, log=None)
        handler = get_ingestion_processor(ingestion_type)
        result = handler(submission)
        if DEBUG_SUBMISSIONS:
            PRINT("DONE PROCESSING FOR %s" % email)
        return {
            "result": result,
            "ingestion_type": ingestion_type,
            "submission_id": submission_id,
        }


@view_config(route_name='queue_ingestion', request_method='POST', permission='index')
@debug_log
def queue_ingestion(context, request):
    """ Queues uuids as part of the request body for ingestion. Can batch as many as desired in a
        single request. Note that you can also pass ingestion_type, which will apply to all uuids queued.
        The default is (SNV) ontology.
    """
    ignored(context)
    uuids = request.json.get('uuids', [])
    ingestion_type = request.json.get('ingestion_type', 'ontology')  # note that this applies to all uuids
    override_name = request.json.get('override_name', None)
    return enqueue_uuids_for_request(request, uuids, ingestion_type=ingestion_type, override_name=override_name)


def enqueue_uuids_for_request(request, uuids, *, ingestion_type, override_name=None):
    response = {
        'notification': 'Failure',
        'number_queued': 0,
        'detail': 'Nothing was queued. Make sure to past in a list of uuids in in "uuids" key.'
    }
    if uuids is []:
        return response
    queue_manager = get_queue_manager(request, override_name=override_name)
    _, failed = queue_manager.add_uuids(uuids, ingestion_type=ingestion_type)
    if not failed:
        response['notification'] = 'Success'
        response['number_queued'] = len(uuids)
        response['detail'] = 'Successfully queued the following uuids: %s' % uuids
        app_project().note_ingestion_enqueue_uuids_for_request(ingestion_type, request, uuids)
    else:
        response['number_queued'] = len(uuids) - len(failed)
        response['detail'] = 'Some uuids failed: %s' % failed
    return response


def get_queue_manager(request, *, override_name):
    return (request.registry[INGESTION_QUEUE]
            if not override_name
            else IngestionQueueManager(request.registry, override_name=override_name))


class IngestionListener(IngestionListenerBase):
    """ Organizes helper functions for the ingestion listener """
    POLL_INTERVAL = 10  # seconds between each poll
    INGEST_AS_USER = environ_bool('INGEST_AS_USER', default=True)  # The new way, but possible to disable for now

    def __init__(self, vapp, _queue_manager=None, _update_status=None):
        self.vapp = vapp

        # Get queue_manager
        registry = None
        if isinstance(self.vapp, (webtest.TestApp, VirtualApp)):  # TestApp in testing or VirtualApp in production
            registry = self.vapp.app.registry
        elif _queue_manager is None:  # if we got here, we cannot succeed in starting
            raise Exception('Bad arguments given to IngestionListener: %s, %s, %s' %
                            (self.vapp, _queue_manager, _update_status))
        self.queue_manager = IngestionQueueManager(registry) if not _queue_manager else _queue_manager
        self.update_status = _update_status

    @staticmethod
    def should_remain_online(override=None):
        """ A function that says whether 'run' should continue. This is provided because it
            can be mocked in testing.

            :param override: a lambda that will execute when evaluating if specified
            :return: True if should stay running, False otherwise
        """
        if not override:
            return True
        return override()

    def get_messages(self):
        """ Sleeps (as to not hit SQS too frequently) then requests messages,
            returning the result bodies.

            NOTE: THIS FUNCTION SHOULD NOT BE USED OUTSIDE OF THIS CODE SINCE
            IT BLOCKS FOR RATE LIMITING REASONS

        :return: messages available on SQS
        """
        time.sleep(self.POLL_INTERVAL)  # sleep here before polling again
        return self.queue_manager.receive_messages()

    def delete_messages(self, messages):
        """ Deletes messages from SQS (after they have been processed). Does not return
            anything but will log if messages fail deletion.

        :param messages: messages to be deleted
        """
        failed = self.queue_manager.delete_messages(messages)
        while True:
            debuglog("Trying to delete messages")
            tries = 3
            if failed:
                debuglog("Failed to delete messages")
                if tries > 0:
                    failed = self.queue_manager.delete_messages(failed)  # try again
                    tries -= 1
                else:
                    log.error('Failed to delete messages from SQS: %s' % failed)
                    break
            else:
                debuglog("Deleted messages")
                break

    def _patch_value(self, uuid, field, value):
        """ Patches field with value on item uuid """
        self.vapp.patch_json('/' + uuid, {field: value})

    def patch_ingestion_report(self, report, uuid):
        """ Sets the file_ingestion_error field of the given uuid """
        if isinstance(report, IngestionReport):  # handle normal case
            self._patch_value(uuid, 'file_ingestion_error', report.get_errors())
        elif isinstance(report, list):  # handle when build_ingestion_error_report result is passed
            self._patch_value(uuid, 'file_ingestion_error', report)
        else:
            raise TypeError('Got bad type for ingestion error report: %s' % report)

    def set_status(self, uuid, status):
        """ Sets the file_ingestion_status of the given uuid """
        self._patch_value(uuid, 'file_ingestion_status', status)

    @staticmethod
    def build_ingestion_error_report(msg):
        """ Builds an ingestion error report in case an error is encountered that cannot be recovered from
            in VCF ingestion - see file_processed.json for structure definition. """
        return [
            {
                'body': msg,
                'row': -1  # this exception may have occurred on a particular row but since it could not be recovered
            }              # from we assume the msg has sufficient info to work backwards from - Will 4/9/21
        ]

    def run(self):
        """ Main process for this class. Runs forever doing ingestion as needed.

            HIGH LEVEL LOGIC:
                while True:
                    while there are messages available:
                        for each message:
                            download, decompress, ingest, patch file status to "Ingested"
                        delete processed messages
        """
        log.info('Ingestion listener successfully online.')

        debuglog("Ingestion listener started.")

        messages = []  # This'll get a better value below in each loop iteration. This is just a declaration of intent.

        def discard(msg):
            self.delete_messages([msg])
            # Assuming we didn't get an error trying to remove it,
            # it should also get removed from our to-do list.
            messages.remove(msg)

        while self.should_remain_online():

            debuglog("About to get messages.")

            messages = self.get_messages()  # wait here

            debuglog("Got", len(messages), "messages.")

            # ingest each VCF file
            for message in list(messages):
                # C4-990/2023-02-09/dmichaels
                # Added the list wrapper around messages in the above loop,
                # i.e. list(messages), so that when we remove a message from
                # the messages list via the discard function (above) the loop
                # does not end up skipping the very next message.

                debuglog("Message:", message)

                # C4-990/2023-02-09/dmichaels
                # This calls at most one our message handlers
                # registered via the @ingestion_message_handler decorator.
                if call_ingestion_message_handler(message, self):
                    # Here one of our message handlers was called and it processed this message.
                    discard(message)

            # This is just fallback cleanup in case messages weren't cleaned up within the loop.
            # In normal operation, they will be.
            self.delete_messages(messages)


def run(vapp=None, _queue_manager=None, _update_status=None):
    """ Entry-point for the ingestion listener for waitress. """
    ingestion_listener = IngestionListener(vapp, _queue_manager=_queue_manager, _update_status=_update_status)
    try:
        ingestion_listener.run()
    except Exception as e:
        debuglog(str(e))
        raise


class ErrorHandlingThread(threading.Thread):
    """ Must be duplicated here so logging is correct. """

    def run(self):
        # interval = self._kwargs.get('interval', DEFAULT_INTERVAL)
        interval = 60  # DB polling can and should be slower
        update_status = self._kwargs['_update_status']  # noQA - uses private instance variables of parent class
        while True:
            try:
                self._target(*self._args, **self._kwargs)  # noQA - uses private instance variables of parent class
            except (psycopg2.OperationalError, elasticsearch.exceptions.ConnectionError) as e:
                # Handle database restart
                log.warning('Database not there, maybe starting up: %r', e)
                update_status(msg=repr(e))
                log.debug('sleeping')
                time.sleep(interval)
                continue
            except Exception as e:
                # Unfortunately mod_wsgi does not restart immediately
                log.exception('Exception in ingestion listener, restarting process at next request: %s' % e)
                os.kill(os.getpid(), signal.SIGINT)
            break


# Composite Application (for wsgi)
def composite(loader, global_conf, **settings):
    """ This is a composite pyramid app, meant to run components of an application
        or an application extension. In our case we are running the ingestion listener,
        which requires executing a command with application context. This code lives
        in encoded top-level as it is a wsgi entry-point. Note that the local deployment
        does NOT run the listener this way, but runs the run method through main directly.
        This code is heavily based off of the es_index_listener in snovault.
    """
    listener = None

    # Register before app creation.
    @atexit.register
    def join_listener():
        if listener:
            log.debug('joining listening thread')
            listener.join()

    # Composite app is used so we can load the main app
    app_name = settings.get('app', None)
    app = loader.get_app(app_name, global_conf=global_conf)
    username = settings.get('username', 'IMPORT')
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': username,
    }
    vapp = VirtualApp(app, environ)
    timestamp = datetime.datetime.utcnow().isoformat()
    status_holder = {
        'status': {
            'status': 'starting listener',
            'started': timestamp,
            'msgs': []
        },
    }

    def update_status(msg=None, **kw):
        """ Method passed to run to update "global" status. """
        # Setting a value in a dictionary is atomic
        status = status_holder['status'].copy()
        status.update(**kw)  # can hold generic info
        if msg is not None:
            status['msgs'].append(msg)
        status_holder['status'] = status

    kwargs = {
        'vapp': vapp,
        '_update_status': update_status
    }

    # daemon thread that actually executes `run` method to call /index
    listener = ErrorHandlingThread(target=run, name='listener', kwargs=kwargs)
    listener.daemon = True
    log.debug('WSGI Ingestion Listener Started')
    listener.start()

    # Register after virtualapp creation.
    @atexit.register
    def shutdown_listener():
        """ Echo a statement at shutdown """
        log.debug('shutting down listening thread')

    def status_app(environ, start_response):
        """ Allows you to get the status of the ingestion "manager". This will be much
            more useful once multi-processing is thrown at ingestion.
        """
        ignored(environ)
        status = '200 OK'
        response_headers = [('Content-type', 'application/json')]
        start_response(status, response_headers)
        return [json.dumps(status_holder['status'])]

    return status_app


# Command Application (for waitress)
def main():
    """ Entry point for the local deployment. """
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description='Listen for VCF File uuids to ingest',
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('--app-name', help='Pyramid app name in configfile')
    parser.add_argument('--username', '-u', default='IMPORT', help='Import username')
    parser.add_argument('--dry-run', action='store_true', help='Do not post variants, just validate')
    parser.add_argument('config_uri', help="path to configfile")
    args = parser.parse_args()

    app = paster.get_app(args.config_uri, args.app_name)
    config = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': args.username,
    }

    vapp = VirtualApp(app, config)
    return run(vapp)


if __name__ == '__main__':
    main()



import venusian

from .interfaces import TYPES, UPGRADER
from pkg_resources import parse_version as parse_pkg_version
from pyramid.interfaces import PHASE1_CONFIG
from .interfaces import PHASE2_5_CONFIG


# These are weird values for defaults, but they are backward-compatible and this code is very tricky. -kmp 6-Feb-2023
DEFAULT_VERSION_STRING = ''
IMPOSSIBLY_LOW_VERSION_STRING = '0'


def parse_version(version_string: str):
    # The expectation is that version_string will represent a version equal to or greater than '1',
    # except that if the default version string ('') is given, it will represent a version
    # that is less than all other version strings, IMPOSSIBLY_LOW_VERSION_STRING ('0').
    return parse_pkg_version(version_string or IMPOSSIBLY_LOW_VERSION_STRING)


DEFAULT_VERSION = parse_version(DEFAULT_VERSION_STRING)


def includeme(config):
    config.include('.typeinfo')
    config.registry[UPGRADER] = Upgrader(config.registry)
    config.add_directive('add_upgrade', add_upgrade)
    config.add_directive('add_upgrade_step', add_upgrade_step)
    config.add_directive('set_upgrade_finalizer', set_upgrade_finalizer)
    config.add_directive(
        'set_default_upgrade_finalizer', set_default_upgrade_finalizer)
    config.add_request_method(upgrade, 'upgrade')


class ConfigurationError(Exception):
    pass


class UpgradeError(Exception):
    pass


class UpgradePathNotFound(UpgradeError):
    def __str__(self):
        return "%r from %r to %r (at %r)" % self.args


class VersionTooHigh(UpgradeError):
    pass


class Upgrader(object):
    """ Upgrade manager
    """
    def __init__(self, registry):
        self.types = registry[TYPES]
        self.schema_upgraders = {}
        self.default_finalizer = None

    def add_upgrade(self, schema_name, version, finalizer=None):
        if schema_name in self.schema_upgraders:
            raise ConfigurationError('duplicate schema_name', schema_name)
        if finalizer is None:
            finalizer = self.default_finalizer
        schema_upgrader = SchemaUpgrader(schema_name, version, finalizer)
        self.schema_upgraders[schema_name] = schema_upgrader

    def upgrade(self, type_, value,
                current_version='1', target_version=None, **kw):
        schema_name = self.types[type_].name
        schema_upgrader = self.schema_upgraders[schema_name]
        return schema_upgrader.upgrade(
            value, current_version, target_version, **kw)

    def __getitem__(self, schema_name):
        return self.schema_upgraders[schema_name]

    def __contains__(self, schema_name):
        return schema_name in self.schema_upgraders


class SchemaUpgrader(object):
    """ Manages upgrade steps
    """
    def __init__(self, name, version, finalizer=None):
        self.__name__ = name
        self.version = version
        self.upgrade_steps = {}
        self.finalizer = finalizer

    def add_upgrade_step(self, step, source=DEFAULT_VERSION_STRING, dest=None):
        if dest is None:
            dest = self.version
        if parse_version(dest) <= parse_version(source):
            raise ValueError(f"In add_upgrade_step, the dest {dest!r} is not greater than source {source!r}.")
        if parse_version(source) in self.upgrade_steps:
            raise ConfigurationError(f"In add_upgrade_step, a duplicate step was given for source {source!r}.")
        self.upgrade_steps[parse_version(source)] = UpgradeStep(step, source, dest)

    def upgrade(self, value, current_version='1', target_version=None, **kw):
        if target_version is None:
            target_version = self.version

        if parse_version(current_version) > parse_version(target_version):
            raise VersionTooHigh(self.__name__, current_version, target_version)

        # Try to find a path from current to target versions
        steps = []
        version = current_version

        # If no entry exists for the current_version, fallback to DEFAULT_VERSION_STRING
        if parse_version(version) not in self.upgrade_steps:
            try:
                step = self.upgrade_steps[DEFAULT_VERSION]
            except KeyError:
                pass
            else:
                if parse_version(step.dest) >= parse_version(version):
                    steps.append(step)
                    version = step.dest

        while parse_version(version) < parse_version(target_version):
            try:
                step = self.upgrade_steps[parse_version(version)]
            except KeyError:
                break
            steps.append(step)
            version = step.dest

        if version != target_version:
            raise UpgradePathNotFound(
                self.__name__, current_version, target_version, version)

        # Apply the steps

        system = {}
        system.update(kw)

        for step in steps:
            next_value = step(value, system)
            if next_value is not None:
                value = next_value

        if self.finalizer is not None:
            next_value = self.finalizer(value, system, version)
            if next_value is not None:
                value = next_value

        return value


class UpgradeStep(object):
    def __init__(self, step, source, dest):
        self.step = step
        self.source = source
        self.dest = dest

    def __call__(self, value, system):
        return self.step(value, system)


# Imperative configuration

def add_upgrade(config, type_, version, finalizer=None):
    if finalizer is not None:
        config.set_upgrade_finalizer(type_, finalizer)

    def callback():
        types = config.registry[TYPES]
        ti = types[type_]
        upgrader = config.registry[UPGRADER]
        upgrader.add_upgrade(ti.name, version)

    config.action(
        ('add_upgrade', type_),
        callback, order=PHASE2_5_CONFIG)


def add_upgrade_step(config, type_, step, source=DEFAULT_VERSION_STRING, dest=None):

    def callback():
        types = config.registry[TYPES]
        ti = types[type_]
        upgrader = config.registry[UPGRADER]
        upgrader[ti.name].add_upgrade_step(step, source, dest)

    config.action(
        ('add_upgrade_step', type_, parse_version(source)),
        callback)


def set_upgrade_finalizer(config, type_, finalizer):

    def callback():
        types = config.registry[TYPES]
        ti = types[type_]
        upgrader = config.registry[UPGRADER]
        upgrader[ti.name].finalizer = finalizer

    config.action(
        ('set_upgrade_finalizer', type_),
        callback)


def set_default_upgrade_finalizer(config, finalizer):

    def callback():
        upgrader = config.registry[UPGRADER]
        upgrader.default_finalizer = finalizer

    config.action(
        'set_default_upgrade_finalizer',
        callback, order=PHASE1_CONFIG)


# Declarative configuration

def upgrade_step(schema_name, source=DEFAULT_VERSION_STRING, dest=None):
    """ Register an upgrade step
    """

    def decorate(step):
        def callback(scanner, factory_name, factory):
            scanner.config.add_upgrade_step(schema_name, step, source, dest)

        venusian.attach(step, callback, category='snovault.upgrader')
        return step

    return decorate


def upgrade_finalizer(schema_name):
    """ Register a finalizer
    """

    def decorate(finalizer):
        def callback(scanner, factory_name, factory):
            scanner.config.set_upgrade_finalizer(schema_name, finalizer)

        venusian.attach(finalizer, callback, category='snovault.upgrader')
        return finalizer

    return decorate


def default_upgrade_finalizer(finalizer):
    def callback(scanner, factory_name, factory):
        scanner.config.set_default_upgrade_finalizer(finalizer)

    venusian.attach(finalizer, callback, category='snovault.upgrader')
    return finalizer


# Upgrade
def upgrade(request, schema_name, value,
            current_version='1', target_version=None, **kw):
    upgrader = request.registry[UPGRADER]
    return upgrader.upgrade(
        schema_name, value, current_version=current_version, target_version=None,
        request=request, **kw)



# -*- coding: utf-8 -*-
"""Load collections and determine the order."""

from base64 import b64encode
import gzip
import json
import magic
import mimetypes
import os
from PIL import Image
import re
import structlog
import traceback
from typing import List, Optional, Tuple, Union
from webtest import TestApp
from webtest.response import TestResponse as TestAppResponse
import uuid
from pyramid.paster import get_app
from pyramid.response import Response
from pyramid.router import Router
from pyramid.view import view_config
from dcicutils.data_readers import RowReader
from dcicutils.misc_utils import ignored, environ_bool, to_camel_case, VirtualApp
from dcicutils.secrets_utils import assume_identity
from snovault.util import debug_log
from .schema_utils import get_identifying_and_required_properties
from .project_app import app_project
from .server_defaults_misc import add_last_modified


text = type(u'')
logger = structlog.getLogger(__name__)


def includeme(config):
    # provide an endpoint to do bulk uploading that just uses loadxl
    config.add_route('load_data', '/load_data')
    config.scan(__name__)


IS_ATTACHMENT = [
    'attachment',
    'file_format_specification',
]


# This uuid should be constant across all portals
LOADXL_USER_UUID = "3202fd57-44d2-44fb-a131-afb1e43d8ae5"


# order of items references with linkTo in a field in  'required' in schemas
def loadxl_order():
    # This should be set by the downstream application
    return app_project().loadxl_order()


class LoadGenWrapper(object):
    """
    Simple class that accepts a generator function and handles errors by
    setting self.caught to the error message.
    """
    def __init__(self, gen):
        self.gen = gen
        self.caught = None

    def __iter__(self):
        """
        Iterate through self.gen and see if 'ERROR: ' bytes are in any yielded
        value. If so, store the error message as self.caught and raise
        StopIteration to halt the generator.
        """
        # self.caught = yield from self.gen
        for iter_val in self.gen:
            if b'ERROR:' in iter_val:
                self.caught = iter_val.decode()
            yield iter_val

    def close(self):
        if self.caught:
            logger.error('load_data: failed to load with iter_response', error=self.caught)


@view_config(route_name='load_data', request_method='POST', permission='add')
@debug_log
def load_data_view(context, request):
    """
    expected input data

    {'local_path': path to a directory or file in file system
     'fdn_dir': inserts folder under encoded/tests/data
     'store': if not local_path or fdn_dir, look for a dictionary of items here
     'overwrite' (Bool): overwrite if existing data
     'itype': (list or str): only pick some types from the source or specify type in in_file
     'iter_response': invoke the Response as an app_iter, directly calling load_all_gen
     'config_uri': user supplied configuration file}

    post can contain 2 different styles of data
    1) reference to a folder or file (local_path or fd_dir). If this is done
       itype can be optionally used to specify type of items loaded from files
    2) store in form of {'item_type': [items], 'item_type2': [items]}
       item_type should be same as insert file names i.e. file_fastq
    """
    ignored(context)
    # this is a bit weird but want to reuse load_data functionality so I'm rolling with it
    config_uri = request.json.get('config_uri', 'production.ini')
    patch_only = request.json.get('patch_only', False)
    post_only = request.json.get('post_only', False)
    testapp = create_testapp(config_uri)
    # expected response
    request.response.status = 200
    result = {
        'status': 'success',
        '@type': ['result'],
    }
    store = request.json.get('store', {})
    local_path = request.json.get('local_path')
    fdn_dir = request.json.get('fdn_dir')
    overwrite = request.json.get('overwrite', False)
    itype = request.json.get('itype')
    iter_resp = request.json.get('iter_response', False)
    inserts = None
    from_json = False
    if fdn_dir:
        inserts = app_project().project_filename(os.path.join('tests/data/', fdn_dir) + '/')
    elif local_path:
        inserts = local_path
    elif store:
        inserts = store
        from_json = True
    # if we want to iterate over the response to keep the connection alive
    # this directly calls load_all_gen, instead of load_all
    if iter_resp:
        return Response(
            content_type='text/plain',
            app_iter=LoadGenWrapper(
                load_all_gen(testapp, inserts, None, overwrite=overwrite, itype=itype,
                             from_json=from_json, patch_only=patch_only, post_only=post_only)
            )
        )
    # otherwise, it is a regular view and we can call load_all as usual
    if inserts:
        res = load_all(testapp, inserts, None, overwrite=overwrite, itype=itype, from_json=from_json)
    else:
        res = 'No uploadable content found!'

    if res:  # None if load_all is successful
        print(LOAD_ERROR_MESSAGE)
        request.response.status = 422
        result['status'] = 'error'
        result['@graph'] = str(res)
    return result


def trim(value):
    """Shorten excessively long fields in error log."""
    if isinstance(value, dict):
        return {k: trim(v) for k, v in value.items()}
    if isinstance(value, list):
        return [trim(v) for v in value]
    if isinstance(value, str) and len(value) > 160:
        return value[:77] + '...' + value[-80:]
    return value


def find_doc(docsdir, filename):
    """tries to find the file, if not returns false."""
    path = None
    if not docsdir:
        return
    for dirpath in docsdir:
        candidate = os.path.join(dirpath, filename)
        if not os.path.exists(candidate):
            continue
        if path is not None:
            msg = 'Duplicate filenames: %s, %s' % (path, candidate)
            raise ValueError(msg)
        path = candidate
    if path is None:
        return
    return path


def attachment(path):
    """Create an attachment upload object from a filename Embeds the attachment as a data url."""
    filename = os.path.basename(path)
    mime_type, encoding = mimetypes.guess_type(path)
    major, minor = mime_type.split('/')
    try:
        detected_type = magic.from_file(path, mime=True).decode('ascii')
    except AttributeError:
        detected_type = magic.from_file(path, mime=True)
    # XXX This validation logic should move server-side.
    if not (detected_type == mime_type or
            detected_type == 'text/plain' and major == 'text'):
        raise ValueError('Wrong extension for %s: %s' % (detected_type, filename))
    with open(path, 'rb') as stream:
        attach = {'download': filename,
                  'type': mime_type,
                  'href': 'data:%s;base64,%s' % (mime_type, b64encode(stream.read()).decode('ascii'))}
        if mime_type in ('application/pdf', "application/zip", 'text/plain',
                         'text/tab-separated-values', 'text/html', 'application/msword', 'application/vnd.ms-excel',
                         'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'):
            # XXX Should use chardet to detect charset for text files here.
            return attach
        if major == 'image' and minor in ('png', 'jpeg', 'gif', 'tiff'):
            # XXX we should just convert our tiffs to pngs
            stream.seek(0, 0)
            im = Image.open(stream)
            im.verify()
            if im.format != minor.upper():
                msg = "Image file format %r does not match extension for %s"
                raise ValueError(msg % (im.format, filename))
            attach['width'], attach['height'] = im.size
            return attach
    raise ValueError("Unknown file type for %s" % filename)


def format_for_attachment(json_data, docsdir):
    for field in IS_ATTACHMENT:
        if field in json_data:
            if isinstance(json_data[field], dict):
                pass
            elif isinstance(json_data[field], str):
                path = find_doc(docsdir, json_data[field])
                if not path:
                    del json_data[field]
                    logger.error(f'Removing {field} form {json_data["uuid"]}, expecting path')
                else:
                    json_data[field] = attachment(path)
            else:
                # malformatted attachment
                del json_data[field]
                logger.error(f'Removing {field} form {json_data["uuid"]}, expecting path')
    return json_data


LOAD_ERROR_MESSAGE = """#   ██▓     ▒█████   ▄▄▄      ▓█████▄  ██▓ ███▄    █   ▄████
#  ▓██▒    ▒██▒  ██▒▒████▄    ▒██▀ ██▌▓██▒ ██ ▀█   █  ██▒ ▀█▒
#  ▒██░    ▒██░  ██▒▒██  ▀█▄  ░██   █▌▒██▒▓██  ▀█ ██▒▒██░▄▄▄░
#  ▒██░    ▒██   ██░░██▄▄▄▄██ ░▓█▄   ▌░██░▓██▒  ▐▌██▒░▓█  ██▓
#  ░██████▒░ ████▓▒░ ▓█   ▓██▒░▒████▓ ░██░▒██░   ▓██░░▒▓███▀▒
#  ░ ▒░▓  ░░ ▒░▒░▒░  ▒▒   ▓▒█░ ▒▒▓  ▒ ░▓  ░ ▒░   ▒ ▒  ░▒   ▒
#  ░ ░ ▒  ░  ░ ▒ ▒░   ▒   ▒▒ ░ ░ ▒  ▒  ▒ ░░ ░░   ░ ▒░  ░   ░
#    ░ ░   ░ ░ ░ ▒    ░   ▒    ░ ░  ░  ▒ ░   ░   ░ ░ ░ ░   ░
#      ░  ░    ░ ░        ░  ░   ░     ░           ░       ░
#                              ░
#   ██▓ ███▄    █   ██████ ▓█████  ██▀███  ▄▄▄█████▓  ██████
#  ▓██▒ ██ ▀█   █ ▒██    ▒ ▓█   ▀ ▓██ ▒ ██▒▓  ██▒ ▓▒▒██    ▒
#  ▒██▒▓██  ▀█ ██▒░ ▓██▄   ▒███   ▓██ ░▄█ ▒▒ ▓██░ ▒░░ ▓██▄
#  ░██░▓██▒  ▐▌██▒  ▒   ██▒▒▓█  ▄ ▒██▀▀█▄  ░ ▓██▓ ░   ▒   ██▒
#  ░██░▒██░   ▓██░▒██████▒▒░▒████▒░██▓ ▒██▒  ▒██▒ ░ ▒██████▒▒
#  ░▓  ░ ▒░   ▒ ▒ ▒ ▒▓▒ ▒ ░░░ ▒░ ░░ ▒▓ ░▒▓░  ▒ ░░   ▒ ▒▓▒ ▒ ░
#   ▒ ░░ ░░   ░ ▒░░ ░▒  ░ ░ ░ ░  ░  ░▒ ░ ▒░    ░    ░ ░▒  ░ ░
#   ▒ ░   ░   ░ ░ ░  ░  ░     ░     ░░   ░   ░      ░  ░  ░
#   ░           ░       ░     ░  ░   ░                    ░
#
#    █████▒▄▄▄       ██▓ ██▓    ▓█████ ▓█████▄
#  ▓██   ▒▒████▄    ▓██▒▓██▒    ▓█   ▀ ▒██▀ ██▌
#  ▒████ ░▒██  ▀█▄  ▒██▒▒██░    ▒███   ░██   █▌
#  ░▓█▒  ░░██▄▄▄▄██ ░██░▒██░    ▒▓█  ▄ ░▓█▄   ▌
#  ░▒█░    ▓█   ▓██▒░██░░██████▒░▒████▒░▒████▓
#   ▒ ░    ▒▒   ▓▒█░░▓  ░ ▒░▓  ░░░ ▒░ ░ ▒▒▓  ▒
#   ░       ▒   ▒▒ ░ ▒ ░░ ░ ▒  ░ ░ ░  ░ ░ ▒  ▒
#   ░ ░     ░   ▒    ▒ ░  ░ ░      ░    ░ ░  ░
#               ░  ░ ░      ░  ░   ░  ░   ░
#                                       ░                    """


def load_all(testapp, inserts, docsdir, overwrite=True, itype=None, from_json=False, patch_only=False, post_only=False,
             skip_types=None):
    """
    Wrapper function for load_all_gen, which invokes the generator returned
    from that function. Takes all of the same args as load_all_gen, so
    please reference that docstring.

    This function uses LoadGenWrapper, which will catch a returned value from
    the execution of the generator, which is an Exception in the case of
    load_all_gen. Return that Exception if encountered, which is consistent
    with the functionality of load_all_gen.
    """
    gen = LoadGenWrapper(
        load_all_gen(testapp, inserts, docsdir, overwrite, itype, from_json, patch_only, post_only, skip_types)
    )
    # run the generator; don't worry about the output
    for _ in gen:
        pass
    # gen.caught is None for success and an error message on failure
    if gen.caught is None:
        return None
    else:
        return Exception(gen.caught)


LOADXL_ALLOW_NONE = environ_bool("LOADXL_ALLOW_NONE", default=True)

def get_identifying_properties(item: dict, identifying_properties: list) -> List[str]:
    """
    Returns the list of all identifying properties of the given item which have a value,
    given its identifying properties (passed in from the identifyingProperties of the item's
    type schema). Favor uuid property and defavor aliases; no other ordering defined.
    """
    results = []
    if item.get("uuid"):
        results.append("uuid")
    for identifying_property in identifying_properties:
        if identifying_property not in ["aliases", "uuid"] and item.get(identifying_property) is not None:
            results.append(identifying_property)
    if "aliases" in identifying_properties and item.get("aliases") is not None:
        results.append("aliases")
    return results


def get_identifying_paths(item: dict, item_type: str, identifying_properties: list) -> List[str]:
    """
    Returns the Portal URL path for the first identifying property of the given item which has a
    value, given its identifying properties (passed in from the identifyingProperties of the item's
    type schema). Favor uuid property; no ordering defined for other identifying properties.
    """
    results = []
    for identifying_property in get_identifying_properties(item, identifying_properties):
        if (identifying_value := item.get(identifying_property)):
            if isinstance(identifying_value, list):
                for identifying_value_item in identifying_value:
                    results.append(f"/{item_type}/{identifying_value_item}")
            elif identifying_property == "uuid":
                results.append(f"/{identifying_value}")
            else:
                results.append(f"/{item_type}/{identifying_value}")
    return results


def get_identifying_property(item: dict, identifying_properties: list) -> Optional[str]:
    """
    Returns the first identifying property of the given item which has a value, given its
    identifying properties (passed in from the identifyingProperties of the item's type
    schema). Favor uuid property and defavor aliases; no other ordering defined.
    """
    identifying_properties = get_identifying_properties(item, identifying_properties)
    return identifying_properties[0] if len(identifying_properties) > 0 else None


def get_identifying_path(item: dict, item_type: str, identifying_properties: list) -> Optional[str]:
    identifying_paths = get_identifying_paths(item, item_type, identifying_properties)
    return identifying_paths[0] if identifying_paths else None


def get_identifying_value(item: dict, identifying_properties: list) -> Optional[str]:
    """
    Returns the value of the first identifying property of the given item which has a value,
    given its identifying properties (passed in from the identifyingProperties of the item's
    type schema). Favor uuid property and defavor aliases; no other ordering defined.
    """
    identifying_property = get_identifying_property(item, identifying_properties)
    if identifying_property:
        identifying_property_value = item.get(identifying_property)
        if isinstance(identifying_property_value, list) and len(identifying_property_value) > 0:
            return identifying_property_value[0]
        return identifying_property_value
    return None

def item_exists(app, an_item: dict, a_type: str, identifying_properties: list) -> Optional[str]:
    identifying_paths = get_identifying_paths(an_item, a_type, identifying_properties)
    if not identifying_paths:
        raise Exception("Item has no uuid nor any other identifying property; cannot GET.")
    for identifying_path in identifying_paths:
        try:
            # 301 because @id is the existing item path, not uuid
            # TODO: We we be able to do this part of the process more efficiently in bulk, up front.
            # ALSO: If we could use, for example, "/file-formats/vcf_gz/" (where the trailing slash
            # is important) rather than "/file_format/vcf_gz", we don't get a redirect (301) be get
            # a direct (200) response; but transformation of "file_format" to "file-formats" isn't
            # readily available to us here; and at least in this example (file_format), using the
            # uuid does not get us a direct (200) response, but rather a redirect (301); just FYI.
            existing_item = app.get(identifying_path, status=[200, 301])
            # If we get here then the item exists.
            existing_uuid = get_response_uuid(existing_item)
            return existing_uuid or get_identifying_value(an_item, identifying_properties)
        except Exception:
            # Ignoring exception here on purpose; this happens if the
            # item does not (yet) exist in the database, which is fine.
            pass
    return None


def get_response_uuid(response: TestAppResponse) -> Optional[str]:
    if not response:
        return None
    if response.status_code == 301:
        response = response.follow()
    return response.json.get("uuid") or response.json.get("@graph", [{}])[0].get("uuid")


def normalize_deleted_properties(data: dict) -> Tuple[dict, List[str]]:
    normalized_data = {}
    deleted_properties = []
    # TODO: This is not doing it recursively ...
    for property_name, property_value in data.items():
        if property_value == RowReader.CELL_DELETION_SENTINEL:
            deleted_properties.append(property_name)
        else:
            normalized_data[property_name] = property_value
    return normalized_data, deleted_properties


def load_all_gen(testapp, inserts, docsdir, overwrite=True, itype=None, from_json=False,
                 patch_only=False, post_only=False, skip_types=None, validate_only=False,
                 continue_on_exception: bool = False, verbose=False):
    """
    Generator function that yields bytes information about each item POSTed/PATCHed.
    Is the base functionality of load_all function.

    convert data to store format dictionary (same format expected from from_json=True),
    assume main function is to load reasonable number of inserts from a folder

    Args:
        testapp
        inserts : either a folder, file, or a dictionary in the store format
        docsdir : attachment folder
        overwrite (bool)   : if the database contains the item already, skip or patch
        itype (list or str): limit selection to certain type/types
        from_json (bool)   : if set to true, inserts should be dict instead of folder name
        patch_only (bool)  : if set to true will only do second round patch - no posts
        post_only (bool)   : if set to true posts full item no second round or lookup -
                             use with care - will not work if linkTos to items not in db yet
        skip_types (list)  : if set to a list of item files the process will ignore these files
    Yields:
        Bytes with information on POSTed/PATCHed items

    Returns:
        None if successful, otherwise a bytes error message
    """
    if docsdir is None:
        docsdir = []
    # Collect Items
    store = {}
    if from_json:  # we are directly loading json
        store = inserts
    if not from_json:  # we are loading a file
        use_itype = False
        if os.path.isdir(inserts):  # we've specified a directory
            if not inserts.endswith('/'):
                inserts += '/'
            files = [i for i in os.listdir(inserts) if (i.endswith('.json') or i.endswith('.json.gz'))
                     and (i not in skip_types if skip_types else True)]
        elif os.path.isfile(inserts):  # we've specified a single file
            files = [inserts]
            # use the item type if provided AND not a list
            # otherwise guess from the filename
            use_itype = True if (itype and isinstance(itype, str)) else False
        else:  # cannot get the file
            err_msg = 'Failure loading inserts from %s. Could not find matching file or directory.' % inserts
            print(err_msg)
            yield str.encode(f'ERROR: {err_msg}\n')
            return
            # raise StopIteration
        # load from the directory/file
        for a_file in files:
            if use_itype:
                item_type = itype
            else:
                item_type = a_file.split('/')[-1].split(".")[0]
                a_file = inserts + a_file
            store[item_type] = get_json_file_content(a_file)

    # if there is a defined set of items, subtract the rest
    if itype:
        if isinstance(itype, list):
            store = {i: store[i] for i in itype if i in store}
        else:
            store = {itype: store.get(itype, [])}
    # clear empty values
    store = {k: v for k, v in store.items() if v is not None}
    if not store:
        if LOADXL_ALLOW_NONE:
            return
        if from_json:
            err_msg = 'No items found in input "store" json'
        else:
            err_msg = 'No items found in %s' % inserts
        if itype:
            err_msg += ' for item type(s) %s' % itype
        print(err_msg)
        yield str.encode(f'ERROR: {err_msg}')
        return
        # raise StopIteration
    # order Items
    all_types = list(store.keys())
    for ref_item in reversed(loadxl_order()):
        if ref_item in all_types:
            all_types.insert(0, all_types.pop(all_types.index(ref_item)))
    # collect schemas
    profiles = testapp.get('/profiles/?frame=raw').json

    def get_schema_info(type_name: str) -> (list, list):
        """
        Returns a tuple containing (first) the list of identifying properties and (second) the list
        of any required properties specified by the schema associated with the object of the given
        object type name. The schema is ASSUMED to be contained within the outer profiles dictionary
        variable, keyed by the camel-case version of the given object type name, which itself is
        assumed to be the snake-case version of the type name (though okay if already camel-case).
        See get_identifying_and_required_properties for details of how these fields are extracted.
        """
        schema = profiles[to_camel_case(type_name)]
        return get_identifying_and_required_properties(schema)

    # run step1 - if item does not exist, post with minimal metadata (and skip indexing since we will patch
    # in round 2)
    second_round_items = {}
    if not patch_only:
        for a_type in all_types:
            first_fields = []
            # minimal schema
            identifying_properties, req_fields = get_schema_info(a_type)
            if not post_only:
                ids = identifying_properties
                # some schemas did not include aliases
                if 'aliases' not in ids:
                    ids.append('aliases')
                # file format is required for files, but its usability depends this field
                if a_type in ['file_format', 'experiment_type']:
                    req_fields.append('valid_item_types')
                first_fields = list(set(req_fields + ids))
            skip_existing_items = set()
            posted = 0
            skip_exist = 0
            for an_item in store[a_type]:
                existing_item_identifying_value = None
                if not post_only:
                    existing_item_identifying_value = item_exists(testapp, an_item, a_type, identifying_properties)
                    """
                    identifying_paths = get_identifying_paths(an_item, a_type, identifying_properties)
                    if not identifying_paths:
                        raise Exception("Item has no uuid nor any other identifying property; cannot GET.")
                    for identifying_path in identifying_paths:
                        try:
                            # 301 because @id is the existing item path, not uuid
                            # TODO: We we be able to do this part of the process more efficiently in bulk, up front.
                            # ALSO: If we could use, for example, "/file-formats/vcf_gz/" (where the trailing slash
                            # is important) rather than "/file_format/vcf_gz", we don't get a redirect (301) be get
                            # a direct (200) response; but transformation of "file_format" to "file-formats" isn't
                            # readily available to us here; and at least in this example (file_format), using the
                            # uuid does not get us a direct (200) response, but rather a redirect (301); just FYI.
                            existing_item = testapp.get(identifying_path, status=[200, 301])
                            # If we get here then the item exists.
                            exists = True
                            break
                        except Exception:
                            # Ignoring exception here on purpose; this happens if the
                            # item does not (yet) exist in the database, which is fine.
                            pass
                    """

                    """
                    identifying_path = get_identifying_path(an_item, a_type, identifying_properties)
                    if not identifying_path:
                        raise Exception("Item has no uuid nor any other identifying property; cannot GET.")
                    try:
                        # 301 because @id is the existing item path, not uuid
                        # TODO: We we be able to do this part of the process more efficiently in bulk, up front.
                        # ALSO: If we could use, for example, "/file-formats/vcf_gz/" (where the trailing slash
                        # is important) rather than "/file_format/vcf_gz", we don't get a redirect (301) be get
                        # a direct (200) response; but transformation of "file_format" to "file-formats" isn't
                        # readily available to us here; and at least in this example (file_format), using the
                        # uuid does not get us a direct (200) response, but rather a redirect (301); just FYI.
                        existing_item = testapp.get(identifying_path, status=[200, 301])
                        # If we get here then the item exists.
                        exists = True
                    except Exception:
                        # Ignoring exception here on purpose; this happens if the
                        # item does not (yet) exist in the database, which is fine.
                        pass
                    """
                # skip the items that exists
                # if overwrite=True, still include them in PATCH round
                if existing_item_identifying_value:
                    skip_exist += 1
                    identifying_value = existing_item_identifying_value
                    # identifying_value = (get_response_uuid(existing_item) or
                    #                      get_identifying_value(an_item, identifying_properties))
                    if not overwrite:
                        skip_existing_items.add(identifying_value)
                    if verbose and (filename := an_item.get("filename", "")):
                        filename = " " + filename
                    else:
                        filename = ""
                    yield str.encode(f'SKIP: {identifying_value}{" " + a_type if verbose else ""}{filename}\n')
                else:
                    an_item, _ = normalize_deleted_properties(an_item)
                    if post_only:
                        to_post = an_item
                    else:
                        to_post = {key: value for (key, value) in an_item.items() if key in first_fields}
                    post_request = f'/{a_type}?skip_indexing=true'
                    if validate_only:
                        post_request += '&check_only=true'
                    to_post = format_for_attachment(to_post, docsdir)
                    try:
                        # This creates the (as yet non-existent) item to the
                        # database with just the minimal data (first_fields).
                        res = testapp.post_json(post_request, to_post)  # skip indexing in round 1
                        if not validate_only:
                            assert res.status_code == 201
                            posted += 1
                            # yield bytes to work with Response.app_iter
                            uuid = get_response_uuid(res)
                            if uuid and "uuid" not in an_item:
                                an_item["uuid"] = uuid  # update our item with the new uuid; maybe controversial?
                            identifying_value = (uuid or get_identifying_value(an_item, identifying_properties))
                            if verbose and (filename := an_item.get("filename", "")):
                                filename = " " + filename
                            else:
                                filename = ""
                            yield str.encode(f'POST: {identifying_value}{" " + a_type if verbose else ""}{filename}\n')
                        else:
                            assert res.status_code == 200
                            identifying_value = (get_response_uuid(res) or
                                                 get_identifying_value(an_item, identifying_properties))
                            yield str.encode(f'CHECK: {identifying_value}{" " + a_type if verbose else ""}\n')
                    except Exception as e:
                        print('Posting {} failed. Post body:\n{}\nError Message:{}'
                              ''.format(a_type, str(first_fields), str(e)))
                        # remove newlines from error, since they mess with generator output
                        e_str = str(e).replace('\n', '')
                        try:
                            # 2024-02-13: To help out smaht-submitr refererential integrity
                            # checking, include the identifying path of the problematic object
                            message = f"ERROR: {get_identifying_path(an_item, a_type, identifying_properties)} {e_str}"
                        except Exception:
                            message = f"ERROR: {e_str}"
                        yield str.encode(f'{message}\n')
                        if not continue_on_exception:
                            return
                        # raise StopIteration
            if not validate_only:
                if not post_only:
                    second_round_items[a_type] = [i for i in store[a_type]
                                                  if get_identifying_value(i, identifying_properties)
                                                  not in skip_existing_items]
            else:
                second_round_items[a_type] = []
            logger.info('{} 1st: {} items posted, {} items exists.'.format(a_type, posted, skip_exist))
            logger.info('{} 1st: {} items will be patched in second round'
                        .format(a_type, str(len(second_round_items.get(a_type, [])))))
    elif overwrite and not post_only:
        logger.info('Posting round skipped')
        for a_type in all_types:
            second_round_items[a_type] = [i for i in store[a_type]]
            logger.info('{}: {} items will be patched in second round'
                        .format(a_type, str(len(second_round_items.get(a_type, [])))))

    # Round II - patch the rest of the metadata (ensuring to index by not passing the query param)
    rnd = ' 2nd' if not patch_only else ''
    for a_type in all_types:
        identifying_properties, _ = get_schema_info(a_type)
        patched = 0
        if not second_round_items[a_type]:
            logger.info('{}{}: no items to patch'.format(a_type, rnd))
            continue
        for an_item in second_round_items[a_type]:
            an_item = format_for_attachment(an_item, docsdir)
            try:
                add_last_modified(an_item, userid=LOADXL_USER_UUID)
                identifying_path = get_identifying_path(an_item, a_type, identifying_properties)
                if not identifying_path:
                    raise Exception("Item has no uuid nor any other identifying property; cannot PATCH.")
                normalized_item, deleted_properties = normalize_deleted_properties(an_item)
                if deleted_properties:
                    identifying_path += f"?delete_fields={','.join(deleted_properties)}"
                res = testapp.patch_json(identifying_path, normalized_item)
                assert res.status_code == 200
                patched += 1
                # yield bytes to work with Response.app_iter
                identifying_value = (get_response_uuid(res) or
                                     get_identifying_value(an_item, identifying_properties) or
                                     "<unidentified>")
                if verbose and (filename := an_item.get("filename", "")):
                    filename = " " + filename
                else:
                    filename = ""
                yield str.encode(f'PATCH: {identifying_value}{" " + a_type if verbose else ""}{filename}\n')
            except Exception as e:
                print('Patching {} failed. Patch body:\n{}\n\nError Message:\n{}'.format(
                      a_type, str(an_item), str(e)))
                print('Full error: %s' % traceback.format_exc())
                e_str = str(e).replace('\n', '')
                try:
                    # 2024-02-13: To help out smaht-submitr refererential integrity
                    # checking, include the identifying path of the problematic object
                    message = f"ERROR: {get_identifying_path(an_item, a_type, identifying_properties)} {e_str}"
                except Exception:
                    message = f"ERROR: {e_str}"
                yield str.encode(f'{message}\n')
                if not continue_on_exception:
                    return
                # raise StopIteration
        logger.info('{}{}: {} items patched .'.format(a_type, rnd, patched))

    # explicit return upon finish
    return None


def get_json_file_content(filename):
    """
    Helper function to obtain objects from (compressed) json files.

    :param filename: str file path
    :returns: object loaded from file
    """
    if filename.endswith(".json"):
        with open(filename) as f:
            result = json.loads(f.read())
    elif filename.endswith(".json.gz"):
        with gzip.open(filename) as f:
            result = json.loads(f.read())
    else:
        raise Exception("Expecting a .json or .json.gz file but found %s." % filename)
    return result


def load_data(app, indir='inserts', docsdir=None, overwrite=False,
              use_master_inserts=True, skip_types=None):
    """
    This function will take the inserts folder as input, and place them to the given environment.
    args:
        app:
        indir (inserts): inserts folder, should be relative to tests/data/
        docsdir (None): folder with attachment documents, relative to tests/data
    """
    testapp = create_testapp(app)
    # load master-inserts by default
    if indir != 'master-inserts' and use_master_inserts:
        master_inserts = app_project().project_filename('tests/data/master-inserts/')
        master_res = load_all(testapp, master_inserts, [], skip_types=skip_types)
        if master_res:  # None if successful
            print(LOAD_ERROR_MESSAGE)
            logger.error('load_data: failed to load from %s' % master_inserts, error=master_res)
            return master_res

    if not indir.endswith('/'):
        indir += '/'
    if not os.path.isabs(indir):
        inserts = app_project().project_filename(os.path.join('tests/data/', indir))
    else:
        inserts = indir
    if docsdir is None:
        docsdir = []
    else:
        if not docsdir.endswith('/'):
            docsdir += '/'
        docsdir = [app_project().project_filename(os.path.join('tests/data/', docsdir))]
    res = load_all(testapp, inserts, docsdir, overwrite=overwrite)
    if res:  # None if successful
        print(LOAD_ERROR_MESSAGE)
        logger.error('load_data: failed to load from %s' % docsdir, error=res)
        return res
    return None  # unnecessary, but makes it more clear that no error was encountered


def load_test_data(app, overwrite=False):
    """
    Load inserts and master-inserts

    Returns:
        None if successful, otherwise Exception encountered
    """
    return load_data(app, docsdir='documents', indir='inserts',
                     overwrite=overwrite)


def load_local_data(app, overwrite=False):
    """
    Load inserts from temporary insert folders, if present and populated
    with .json insert files.
    If not present, load inserts and master-inserts.

    Returns:
        None if successful, otherwise Exception encountered
    """

    test_insert_dirs = [
        'temp-local-inserts',
        'demo_inserts'
    ]

    for test_insert_dir in test_insert_dirs:
        chk_dir = app_project().project_filename(os.path.join("tests/data/", test_insert_dir))
        for (dirpath, dirnames, filenames) in os.walk(chk_dir):
            if any([fn for fn in filenames if fn.endswith('.json') or fn.endswith('.json.gz')]):
                logger.info('Loading inserts from "{}" directory.'.format(test_insert_dir))
                return load_data(app, docsdir='documents', indir=test_insert_dir, use_master_inserts=True,
                                 overwrite=overwrite)

    # Default to 'inserts' if no temp inserts found.
    return load_data(app, docsdir='documents', indir='inserts', use_master_inserts=True, overwrite=overwrite)


def load_prod_data(app, overwrite=False):
    """
    Load master-inserts

    Returns:
        None if successful, otherwise Exception encountered
    """
    return load_data(app, indir='master-inserts', overwrite=overwrite)


def load_deploy_data(app, overwrite=True, **kwargs):
    """
    Load deploy-inserts and master-inserts. Overwrites duplicate items
    in both directories to match deploy-inserts version.

    Returns:
        None if successful, otherwise Exception encountered
    """
    return load_data(app, docsdir='documents', indir="deploy-inserts", overwrite=True)


# Set of emails required by the application to function
REQUIRED_USER_CONFIG = [
    {
        'email': 'loadxl@hms.harvard.edu',
        'first_name': 'loadxl',
        'last_name': 'loadxl',
        'uuid': '3202fd57-44d2-44fb-a131-afb1e43d8ae5'
    },
    {
        'email': 'cgap.platform@gmail.com',
        'first_name': 'Platform',
        'last_name': 'Admin',
        'uuid': 'b5f738b6-455a-42e5-bc1c-77fbfd9b15d2'
    },
    {
        'email': 'foursight.app@gmail.com',
        'first_name': 'Foursight',
        'last_name': 'App',
        'uuid': '7677f8a8-79d2-4cff-ab0a-a967a2a68e39'
    },
    {
        'email': 'tibanna.app@gmail.com',
        'first_name': 'Tibanna',
        'last_name': 'App',
        'uuid': 'b041dba8-e2b2-4e54-a621-97edb508a0c4'
    },
]


def load_custom_data(app, overwrite=False):
    """
    Load deploy-inserts and master-inserts, EXCEPT instead of loading the default user.json,
    generate users (if they do not already exist) from the ENCODED_ADMIN_USERS setting in
    the GAC. We assume it has structure consistent with what the template will build in 4dn-cloud-infra
    ie:
        [{"first_name": "John", "last_name": "Doe", "email": "john_doe@example.com"}]
    """
    # start with the users
    testapp = create_testapp(app)
    identity = assume_identity()
    admin_users = json.loads(identity.get('ENCODED_ADMIN_USERS', '{}'))
    if not admin_users:  # we assume you must have set one of these
        print(LOAD_ERROR_MESSAGE)
        logger.error('load_custom_data: failed to load users as none were set - ensure GAC value'
                     ' ENCODED_ADMIN_USERS is set and formatted correctly!')
        return admin_users

    # post all users
    for user in (admin_users + REQUIRED_USER_CONFIG):
        try:
            first_name, last_name, email, _uuid = (user['first_name'], user['last_name'], user['email'],
                                                   user.get('uuid', str(uuid.uuid4())))
        except KeyError:
            print(LOAD_ERROR_MESSAGE)
            logger.error('load_custom_data: failed to load users as they were malformed - ensure GAC value'
                         ' ENCODED_ADMIN_USERS is set, has type array and consists of objects all containing keys'
                         ' and values for first_name, last_name and email!')
            return user
        item = {
            'first_name': first_name,
            'last_name': last_name,
            'email': email,
            'groups': ['admin'],
            'uuid': _uuid
        }
        testapp.post_json('/User', item, status=201)

    res = load_data(app, docsdir='documents', indir='deploy-inserts', overwrite=overwrite, skip_types=['user.json'])
    if res:  # None if successful
        print(LOAD_ERROR_MESSAGE)
        logger.error('load_custom_data: failed to load from deploy-inserts', error=res)
        return res

    return None


def load_cypress_data(app, overwrite=False):
    """
    Load master-inserts and cypress-test-inserts.
    By default, does not overwrite duplicate items in both directories

    Returns:
        None if successful, otherwise Exception encountered
    """
    return load_data(app, indir='cypress-test-inserts', overwrite=overwrite)


def load_data_by_type(app, indir='master-inserts', overwrite=True, itype=None):
    """
    This function will load inserts of type itype from the indir directory.
    args:
        indir (inserts): inserts folder, should be relative to tests/data/
        itype: item type to load (e.g. "higlass_view_config")
    """

    if itype is None:
        print('load_data_by_type: No item type specified. Not loading anything.')
        return

    testapp = create_testapp(app)

    if not indir.endswith('/'):
        indir += '/'
    inserts = app_project().project_filename(os.path.join('tests/data/', indir))

    res = load_all(testapp, inserts, docsdir=[], overwrite=overwrite, itype=itype)
    if res:  # None if successful
        print(LOAD_ERROR_MESSAGE)
        logger.error('load_data_by_type: failed to load from %s' % indir, error=res)
        return res
    return None  # unnecessary, but makes it more clear that no error was encountered


def load_data_via_ingester(vapp: VirtualApp,
                           ontology: dict,
                           itype: Union[str, list] = ["ontology", "ontology_term"],
                           validate_only: bool = False) -> dict:
    """
    Entry point for call from encoded.ingester.processors.handle_ontology_update (2023-03-08).
    Returns dictionary itemizing the created (post), updated (patch), skipped (skip), checked (check),
    and errored (error) ontology term uuids; as well as a count of the number of unique uuids processed;
    the checked category is for validate_only;
    """
    response = load_all_gen(vapp, ontology, None, overwrite=True, itype=itype,
                            from_json=True, patch_only=False, validate_only=validate_only)
    results = {"post": [], "patch": [], "skip": [], "check": [], "error": []}
    unique_uuids = set()
    INGESTION_RESPONSE_PATTERN = re.compile(r"^([A-Z]+): ([0-9a-f-]+)$")
    for item in response:
        # Assume each item in the response looks something like one of (string or bytes):
        # POST: 15425d13-01ce-4e61-be5d-cd04401dff29
        # PATCH: 5b45e66f-7b4f-4923-824b-d0864a689bb
        # SKIP: 4efe24b5-eb17-4406-adb8-060ea2ae2180
        # CHECK: deadbeef-eb17-4406-adb8-0eacafebabe
        # ERROR: 906c4667-483e-4a08-96b9-3ce85ce8bf8c
        # Note that SKIP means skip post/insert; still may to patch/update (if overwrite).
        if isinstance(item, bytes):
            item = item.decode("ascii")
        elif not isinstance(item, str):
            logger.warning(f"load_data_via_ingester: skipping response item of unexpected type ({type(item)}): {item!r}")
            continue
        match = INGESTION_RESPONSE_PATTERN.match(item)
        if not match:
            logger.warning(f"load_data_via_ingester: skipping response item in unexpected form: {item!r}")
            continue
        action = match.group(1).lower()
        uuid = match.group(2)
        if not results.get(action):
            results[action] = []
        results[action].append(uuid)
        unique_uuids.add(uuid)
    results["unique"] = len(unique_uuids)
    return results


def create_testapp(ini_or_app_or_testapp: Union[str, Router, TestApp] = "development.ini",
                   app_name: str = "app") -> TestApp:
    """
    Creates and returns a TestApp; and also adds a get_with_follow method to it.
    Refactored out of above loadxl code (2023-09) to consolidate at a single point,
    and also for use by the generate_local_access_key and view_local_object scripts.
    """
    if isinstance(ini_or_app_or_testapp, TestApp):
        testapp = ini_or_app_or_testapp
    else:
        if isinstance(ini_or_app_or_testapp, Router):
            app = ini_or_app_or_testapp
        else:
            app = get_app(ini_or_app_or_testapp, app_name)
        testapp = TestApp(app, {"HTTP_ACCEPT": "application/json", "REMOTE_USER": "TEST"})
    if not getattr(testapp, "get_with_follow", None):
        def get_with_follow(self, *args, **kwargs):
            raise_exception = kwargs.pop("raise_exception", True)
            if not isinstance(raise_exception, bool):
                raise_exception = True
            try:
                response = self.get(*args, **kwargs)
                if response and response.status_code in [301, 302, 303, 307, 308]:
                    response = response.follow()
                return response
            except Exception as e:
                if raise_exception:
                    raise e
                return None
        testapp.get_with_follow = get_with_follow.__get__(testapp, TestApp)
    return testapp



import json
import structlog
from dcicutils.misc_utils import ignored, PRINT
from .ingestion_listener import (IngestionListener, DEBUG_SUBMISSIONS)
from ..types.ingestion import SubmissionFolio
from ..util import (debuglog, vapp_for_email, make_s3_client)
from .common import metadata_bundles_bucket
from .ingestion_message import IngestionMessage
from .ingestion_message_handler_decorator import ingestion_message_handler
from .ingestion_processor_decorator import get_ingestion_processor


log = structlog.getLogger(__name__)


def includeme(config):
    pass


@ingestion_message_handler(ingestion_type="default")
def ingestion_message_handler_default(message: IngestionMessage, listener: IngestionListener) -> bool:
    """
    This is the part of listener.IngestionListener.run function which handles a
    single message within the (effectively-infinite) incoming message handling loop,
    specifically for non-VCF files; refactored out of ingestion_listener.py February 2023.
    Returns True if the message was successfully handled, otherwise False.
    """

    PRINT(f"Default ingestion message handler called for message ({message.uuid}) type: {message.type}")

    # Let's minimally disrupt things for now. We can refactor this later
    # to make all the parts work the same -kmp
    if listener.INGEST_AS_USER:
        try:
            debuglog("REQUESTING RESTRICTED PROCESSING:", message.uuid)
            process_submission(submission_id=message.uuid,
                               ingestion_type=message.type,
                               # bundles_bucket=submission.bucket,
                               app=listener.vapp.app)
            debuglog("RESTRICTED PROCESSING DONE:", message.uuid)
        except Exception as e:
            log.error(e)
    else:
        submission = SubmissionFolio(vapp=listener.vapp, ingestion_type=message.type,
                                     submission_id=message.uuid)
        handler = get_ingestion_processor(message.type)
        try:
            debuglog("HANDLING:", message.uuid)
            handler(submission)
            debuglog("HANDLED:", message.uuid)
        except Exception as e:
            log.error(e)
    # If we suceeded, we don't need to do it again, and if we failed we don't need to fail again.
    return True


def process_submission(*, submission_id, ingestion_type, app, bundles_bucket=None, s3_client=None):
    ignored(s3_client)  # we might want to restore the ability to pass this, but no one actually does. -kmp 6-Dec-2021
    bundles_bucket = bundles_bucket or metadata_bundles_bucket(app.registry)
    s3_client = make_s3_client()
    manifest_name = "{id}/manifest.json".format(id=submission_id)
    log.warning(f'Processing submission {manifest_name}')
    obj = s3_client.get_object(Bucket=bundles_bucket, Key=manifest_name)
    # data = json.load(obj)['Body']
    data = json.load(obj['Body'])
    email = None
    try:
        email = data['email']
    except KeyError as e:
        ignored(e)
        debuglog("Manifest data is missing 'email' field.")
        if DEBUG_SUBMISSIONS:
            pass
    debuglog("processing submission %s with email %s" % (submission_id, email))
    with vapp_for_email(email=email, app=app) as vapp:
        if DEBUG_SUBMISSIONS:
            PRINT("PROCESSING FOR %s" % email)
        submission = SubmissionFolio(vapp=vapp, ingestion_type=ingestion_type, submission_id=submission_id, log=None)
        handler = get_ingestion_processor(ingestion_type)
        result = handler(submission)
        if DEBUG_SUBMISSIONS:
            PRINT("DONE PROCESSING FOR %s" % email)
        return {
            "result": result,
            "ingestion_type": ingestion_type,
            "submission_id": submission_id,
        }



"""
common.py - tools common to various parts of ingestion
"""

from .exceptions import MissingParameter, BadParameter
from ..util import CONTENT_TYPE_SPECIAL_CASES


def metadata_bundles_bucket(registry):
    return registry.settings.get('metadata_bundles_bucket')


# ==================================================


def register_path_content_type(*, path, content_type):
    """
    Registers that endpoints that begin with the specified path use the indicated content_type.

    This is part of an inelegant workaround for an issue in renderers.py that maybe we can make go away in the future.
    See the 'implementation note' in ingestion/common.py for more details.
    """
    exceptions = CONTENT_TYPE_SPECIAL_CASES.get(content_type, None)
    if exceptions is None:
        CONTENT_TYPE_SPECIAL_CASES[content_type] = exceptions = []
    if path not in exceptions:
        exceptions.append(path)


def content_type_allowed(request):
    """
    Returns True if the current request allows the requested content type.

    This is part of an inelegant workaround for an issue in renderers.py that maybe we can make go away in the future.
    See the 'implementation note' in ingestion/common.py for more details.
    """
    if request.content_type == "application/json":
        # For better or worse, we always allow this.
        return True

    exceptions = CONTENT_TYPE_SPECIAL_CASES.get(request.content_type)

    if exceptions:
        for text in exceptions:
            if text in request.path:
                return True

    return False

# ==================================================


_NO_DEFAULT = object()


def get_parameter(parameter_block, parameter_name, as_type=None, default=_NO_DEFAULT, update=False):
    """
    Returns the value of a given parameter from a dictionary of parameter values.

    If the parameter is not in the dictionary, the default will be returned if one is given.
    If the parameter is not present but there is no default, an error of type MissingParameter will be raised.

    Args:
        parameter_block (dict): a dictionary whose keys are parameter names and whose values are parameter values
        parameter_name (str): the name of a parameter
        as_type: if supplied, a type coercion to perform on the result
        default (object): a default value to be used if the parameter_name is not present.
        update (bool): if as_type is applied, whether to update the parameter_block
    """

    if isinstance(parameter_block, dict):
        if parameter_name in parameter_block:
            parameter_value = parameter_block[parameter_name]
            result = parameter_value
            if as_type:
                if isinstance(as_type, type) and isinstance(result, as_type):
                    return result
                elif as_type is bool:
                    lower_value = str(result).lower()
                    if lower_value == "true":
                        result = True
                    elif lower_value in ("false", "none", "null", ""):
                        result = False
                    else:
                        raise BadParameter(parameter_name=parameter_name, parameter_value=parameter_value,
                                           extra_detail=("Expected a string representing a boolean, such as"
                                                         " 'true' for True, or 'false' or the empty string for False."))
                else:
                    result = as_type(result)
        elif default is _NO_DEFAULT:
            raise MissingParameter(parameter_name=parameter_name)
        else:
            result = default

        if update:
            parameter_block[parameter_name] = result

        return result

    else:
        raise TypeError("Expected parameter_block to be a dict: %s", parameter_block)


class IngestionError:
    """
    Holds info on an error that occurred in ingestion. Right now this consists of the
    offending request body and the VCF row it occurred on.

    This class doesn't really do much except specify the structure. It must align with that of FileProcessed
    (reproduced as of 12/2/2020 below):

        "file_ingestion_error": {
            "title": "Ingestion Error Report",
            "description": "This field is set when an error occurred in ingestion with all errors encountered",
            "type": "array",
            "items": {
                "title": "Ingestion Error",
                "type": "object",
                "properties": {
                    "body": {
                        "type": "string",
                        "index": false  # the intention is not to index this in the future
                    },
                    "row": {
                        "type": "integer"
                    }
                }
            }
        }

    """

    def __init__(self, body, row):
        self.body = body
        self.row = row

    def to_dict(self):
        return {
            'body': str(self.body),
            'row': self.row
        }


class IngestionReport:
    """
    A "virtual" item on file_processed that contains detailed information on the ingestion run.
    Not creating an item for this is a design decision. The output of this process is more for
    debugging and not for auditing, so it does not merit an item at this time.
    """
    MAX_ERRORS = 100  # tune this to get more errors, 100 is a lot though and probably more than needed

    def __init__(self):
        self.grand_total = 0
        self.errors = []

    def brief_summary(self):
        return ('INGESTION REPORT: There were %s total variants detected, of which %s were successful'
                'and %s failed. Check ProcessedFile for full error output.' % (self.grand_total,
                                                                              self.total_successful(),
                                                                              self.total_errors()))

    def total_successful(self):
        return self.grand_total - len(self.errors)

    def total_errors(self):
        return len(self.errors)

    def get_errors(self, limit=True):
        """ Returns a limited number of errors, where limit can be True (self.MAX_ERRORS), False (no limit),
            or an integer. """
        if limit is True:
            limit = self.MAX_ERRORS
        elif limit is False:
            limit = None
        return self.errors[:limit]

    def mark_success(self):
        """ Marks the current row number as successful, which in this case just involves incrementing the total """
        self.grand_total += 1

    def mark_failure(self, *, body, row):
        """ Marks the current row as failed, creating an IngestionError holding the response body and row. """
        self.grand_total += 1
        self.errors.append(IngestionError(body, row).to_dict())



import json
from uuid import (
    UUID,
    uuid4,
)

import transaction
from pyramid.exceptions import HTTPForbidden
from pyramid.settings import asbool
from pyramid.view import view_config
from structlog import get_logger
from dcicutils.diff_utils import DiffManager

from .calculated import calculated_property
from .interfaces import (
    STORAGE,
    Created,
    BeforeModified,
    AfterModified,
)
from .invalidation import add_to_indexing_queue
from .resources import (
    Collection,
    Item,
)
from .util import debug_log
from .validation import ValidationFailure
from .validators import (
    no_validate_item_content_patch,
    no_validate_item_content_post,
    no_validate_item_content_put,
    validate_item_content_patch,
    validate_item_content_post,
    validate_item_content_put,
    validate_item_content_in_place
)

log = get_logger(__name__)


def includeme(config):
    config.include('.calculated')
    config.scan(__name__)


def extract_and_populate_deleted_fields(request, body: dict):
    """ Reads the request parameters looking for 'delete_fields', and if it finds them
        manually add a diff entry for those fields so they are picked up in the diff
        Modifies the body in place
    """
    if 'delete_fields' in request.params:
        deleted_fields = request.params['delete_fields'].split(',')
        for deleted_field in deleted_fields:
            body[deleted_field] = ''


def build_diff_from_request(context, request):
    """ Unpacks request.body as JSON and computes a diff """
    try:
        item_type = context.type_info.name
        body = json.loads(request.body)
        extract_and_populate_deleted_fields(request, body)
        dm = DiffManager(label=item_type)
    except json.decoder.JSONDecodeError:
        log.info('Request body is not valid JSON!')  # can happen from indirect patch, such as with access key
        return None
    except Exception as e:
        log.error('Unknown error encountered building diff from request: %s' % e)
        return None
    return dm.patch_diffs(body)


def create_item(type_info, request, properties, sheets=None):
    '''
    Validates or generates new UUID, instantiates & saves an Item in
    database using provided `type_info` & `properties`, sends 'Created'
    notification, which can be subscribed to using the @subscriber(Created) decorator

    Queues the created item for indexing using a hook on the current transaction
    '''
    registry = request.registry
    item_properties = properties.copy()
    txn = transaction.get()

    if 'uuid' in item_properties:
        try:
            nonvalidated_uuid = item_properties.pop('uuid')
            uuid = UUID(nonvalidated_uuid)
        except ValueError as e:
            raise ValueError(str(e) + ': ' + str(nonvalidated_uuid))
    else:
        uuid = uuid4()

    item = type_info.factory.create(registry, uuid, item_properties, sheets)
    # add the item to the queue
    to_queue = {'uuid': str(item.uuid), 'sid': item.sid}
    telemetry_id = request.params.get('telemetry_id', None)
    if telemetry_id:
        to_queue['telemetry_id'] = telemetry_id
    log.info(event='add_to_indexing_queue', **to_queue)
    txn.addAfterCommitHook(add_to_indexing_queue, args=(request, to_queue, 'add',))
    registry.notify(Created(item, request))

    return item


def update_item(context, request, properties, sheets=None):
    '''
    Updates retrieved-from-database `context` (Item class instance) with
    `properties` (dict) in database, sends 'BeforeModified' & 'AfterModified'
    notifications, which can be subscribed to using the
    @subscriber(BeforeModified) or @subscriber(AfterModified) decorators

    Queues the updated item for indexing using a hook on the current transaction
    '''
    txn = transaction.get()
    registry = request.registry
    item_properties = properties.copy()
    registry.notify(BeforeModified(context, request))
    context.update(item_properties, sheets)
    # set up hook for queueing indexing
    diff = build_diff_from_request(context, request)
    to_queue = {'uuid': str(context.uuid), 'sid': context.sid}
    if diff is not None:
        to_queue['diff'] = diff
    telemetry_id = request.params.get('telemetry_id', None)
    if telemetry_id:
        to_queue['telemetry_id'] = telemetry_id
    txn.addAfterCommitHook(add_to_indexing_queue, args=(request, to_queue, 'edit',))
    registry.notify(AfterModified(context, request))


def delete_item(context, request):
    """
    Sets the status of an item to deleted and triggers indexing
    """
    properties = context.properties.copy()
    properties['status'] = 'deleted'
    update_item(context, request, properties)
    return True


def purge_item(context, request):
    """
    Fully delete an item from the DB and Elasticsearch if all links to that
    have been removed. Requires that the status of the item == 'deleted',
    otherwise will throw a validation failure
    """
    item_type = context.collection.type_info.item_type
    item_uuid = str(context.uuid)
    if context.properties.get('status') != 'deleted':
        msg = (u'Item status must equal deleted before purging from DB.' +
               ' It currently is %s' % context.properties.get('status'))
        raise ValidationFailure('body', ['status'], msg)
    request.registry[STORAGE].purge_uuid(rid=item_uuid, item_type=item_type)
    return True


def render_item(request, context, render, return_uri_also=False):
    if render == 'uuid':
        item_uri = '/%s' % context.uuid
    else:
        item_uri = request.resource_path(context)

    if asbool(render) is True:
        rendered = request.embed(item_uri, '@@object', as_user=True)
    else:
        rendered = item_uri
    return (rendered, item_uri) if return_uri_also else rendered


@view_config(context=Collection, permission='add', request_method='POST',
             validators=[validate_item_content_post])
@view_config(context=Collection, permission='add_unvalidated', request_method='POST',
             validators=[no_validate_item_content_post],
             request_param=['validate=false'])
@debug_log
def collection_add(context, request, render=None):
    '''Endpoint for adding a new Item.'''
    check_only = asbool(request.params.get('check_only', False))
    if check_only:
        return {
            'status': "success",
            '@type': ['result'],
        }

    if render is None:
        render = request.params.get('render', True)

    item = create_item(context.type_info, request, request.validated)

    rendered, item_uri = render_item(request, item, render, True)
    request.response.status = 201
    request.response.location = item_uri
    result = {
        'status': 'success',
        '@type': ['result'],
        '@graph': [rendered],
    }
    return result


@view_config(context=Item, permission='edit', request_method='PUT',
             validators=[validate_item_content_put])
@view_config(context=Item, permission='edit', request_method='PATCH',
             validators=[validate_item_content_patch])
@view_config(context=Item, permission='edit_unvalidated', request_method='PUT',
             validators=[no_validate_item_content_put],
             request_param=['validate=false'])
@view_config(context=Item, permission='edit_unvalidated', request_method='PATCH',
             validators=[no_validate_item_content_patch],
             request_param=['validate=false'])
@view_config(context=Item, permission='index', request_method='GET',
             validators=[validate_item_content_in_place],
             request_param=['check_only=true'])
@debug_log
def item_edit(context, request, render=None):
    '''
    Endpoint for editing an existing Item.
    This handles both PUT and PATCH, difference is the validator

    PUT - replaces the current properties with the new body
    PATCH - updates the current properties with those supplied.
    Note validators will handle the PATH ?delete_fields parameter if you want
    field to be deleted

    Also allow a GET request to just run the edit validators when using
    `check_only=true` query param when indexing. Whether using this or POST/
    PATCH, always return a dummy response with `check_only=true` and do not
    actually modify the DB with `update_item`
    '''
    check_only = asbool(request.params.get('check_only', False))
    if check_only:
        return {
            'status': "success",
            '@type': ['result'],
        }

    if render is None:
        render = request.params.get('render', True)

    # This *sets* the property sheet and adds the item to the indexing queue
    update_item(context, request, request.validated)

    rendered = render_item(request, context, render)
    request.response.status = 200
    result = {
        'status': 'success',
        '@type': ['result'],
        '@graph': [rendered],
    }
    return result


@view_config(context=Item, permission='view', request_method='GET',
             name='links')
@debug_log
def get_linking_items(context, request, render=None):
    """
    Utilize find_uuids_linked_to_item function in PickStorage to find
    any items that link to the given item context
    Split the answer into linkTos and rev_links
    """
    item_uuid = str(context.uuid)
    links = request.registry[STORAGE].find_uuids_linked_to_item(item_uuid)
    request.response.status = 200
    result = {
        'status': 'success',
        '@type': ['result'],
        'display_title': f'Links to {item_uuid}',
        'notification': f'{item_uuid} has {len(links)} items linking to it. This may include rev_links if status != deleted',
        'uuids_linking_to': links
    }
    return result


@view_config(context=Item, permission='edit', request_method='DELETE')
@debug_log
def item_delete_full(context, request, render=None):
    """
    DELETE method that either sets the status of an item to deleted (base
    functionality) or fully purges the item from the DB and ES if all links
    to that item have been removed. Default behavior is delete

    To purge, use ?purge=true query string
    For example: DELETE `/<item-type>/<uuid>?purge=true`
    """
    # possibly temporary fix to check if user is admin
    if hasattr(request, 'user_info'):
        user_details = request.user_info.get('details', {})
    else:
        # used to check for admin here, now done in user_info above
        # note that hasattr() results in a function call when referring to an @property
        user_details = {}
    if 'admin' not in user_details.get('groups', []):
        msg = u'Must be admin to fully delete items.'
        raise ValidationFailure('body', ['userid'], msg)

    purge_from_database = asbool(request.GET and request.GET.get('purge'))
    uuid = str(context.uuid)
    if purge_from_database:
        # Delete entirely - WARNING USE WITH CAUTION - DELETES PERMANENTLY
        # checking of item status and links is done within purge_item()
        result = purge_item(context, request)
        if result:
            return {
                'status': 'success',
                '@type': ['result'],
                'notification': f'Permanently deleted {uuid}',
                '@graph': [uuid]
            }
    else:
        result = delete_item(context, request)
        if result:
            return {
                'status': 'success',
                '@type': ['result'],
                'notification': f'Set status of {uuid} to deleted',
                '@graph': [render_item(request, context, render)]
            }

    return {
        'status': 'failure',
        '@type': ['result'],
        'notification': 'Deletion failed',
        '@graph': [uuid]
    }


@view_config(context=Item, permission='view', request_method='GET',
             name='validation-errors')
@debug_log
def item_view_validation_errors(context, request):
    """
    View config for validation_errors. If the current model is not using ES,
    do not calculate the validation errors as it would require an extra
    request and some tricky permission handling.

    Args:
        context: current Item
        request: current request

    Returns:
        A dictionary including item path and validation errors from ES
    """
    if context.model.used_datastore != 'elasticsearch':
        return {
            '@id': request.resource_path(context),
            'validation_errors': [],
        }
    source = context.model.source
    allowed = set(source['principals_allowed']['view'])  # use view permissions
    if allowed.isdisjoint(request.effective_principals):
        raise HTTPForbidden
    return {
        '@id': source['object']['@id'],
        'validation_errors': source.get('validation_errors', [])
    }


@calculated_property(context=Item, category='page', name='validation-errors',
                     condition=lambda request: request.has_permission('view'))
def validation_errors_property(context, request):
    """
    Frame=page calculated property to add validation_errors to response.
    The request.embed calls item_view_validation_errors
    Args:
        context: current Item
        request: current Request
    Returns:
        List result of validation errors
    """
    path = request.resource_path(context)
    return request.embed(path, '@@validation-errors')['validation_errors']


def get_item_revision_history(request, uuid):
    """ Computes the revision history of the given item from the DB.
        The more edits an item has undergone, the more expensive this
        operation is.
    """
    revisions = request.registry[STORAGE].revision_history(uuid=uuid)
    # Resolve last_modified
    # NOTE: last_modified is a server_default present in our applications that
    # use snovault. This code is intended to resolve the user email of the last_modified
    # resource path (if it exists).
    user_cache = {}
    for revision in revisions:
        last_modified = revision.get('last_modified', {})
        modified_by = last_modified.get('modified_by', None)
        if modified_by:
            user = user_cache.get(modified_by)
            if not user:
                user = request.embed(modified_by, frame='raw')
                user_cache[modified_by] = user
            revision['last_modified']['modified_by'] = user.get('email', 'No email specified!')
    return revisions


@view_config(context=Item, permission='edit', request_method='GET',
             name='revision-history')
@debug_log
def item_view_revision_history(context, request):
    """ View config for viewing an item's revision history.
        For now, to view revision history the caller must have EDIT permissions.
    """
    uuid = str(context.uuid)
    revisions = get_item_revision_history(request, uuid)
    return {
        'uuid': uuid,
        'revisions': revisions
    }



import pytest

from pyramid import paster
from unittest import mock
from ..standalone_dev import make_standalone_app, make_dev_vapp
from .. import standalone_dev as standalone_dev_module

pytestmark = [pytest.mark.working, pytest.mark.unit]


class MockWsgiApp():
    def __init__(self, name, config_uri):
        self.name_for_testing = name
        self.config_uri_for_testing = config_uri


class MockConfigLoader:
    def __init__(self, config_uri):
        self.config_uri = config_uri

    def get_wsgi_app(self, name, options=None):
        assert options is None  # The case we're expecting doesn't pass this argument
        return MockWsgiApp(name=name, config_uri=self.config_uri)


def test_make_standalone_app():

    with mock.patch.object(paster, "get_config_loader", MockConfigLoader):

        with pytest.raises(TypeError):  # positional arguments not allowed
            testapp = make_standalone_app('foo')  # noQA - this is a negative test case

        testapp = make_standalone_app()
        assert testapp.name_for_testing == 'app'
        assert testapp.config_uri_for_testing == 'development.ini'

        testapp = make_standalone_app(name='foo')
        assert testapp.name_for_testing == 'foo'
        assert testapp.config_uri_for_testing == 'development.ini'

        testapp = make_standalone_app(config_file='foo.ini')
        assert testapp.name_for_testing == 'app'
        assert testapp.config_uri_for_testing == 'foo.ini'

        testapp = make_standalone_app(name='foo', config_file='foo.ini')
        assert testapp.name_for_testing == 'foo'
        assert testapp.config_uri_for_testing == 'foo.ini'


class MockVirtualApp:

    def __init__(self, app, environ):
        self.app_for_testing = app
        self.environ_for_testing = environ


def test_make_dev_vapp():

    with mock.patch.object(standalone_dev_module, "VirtualApp", MockVirtualApp):
        with mock.patch.object(paster, "get_config_loader", MockConfigLoader):

            vapp = make_dev_vapp(app='something')
            assert vapp.app_for_testing == 'something'  # If we give something, it's just passed straight through

            vapp = make_dev_vapp()
            assert isinstance(vapp, MockVirtualApp)
            assert isinstance(vapp.app_for_testing, MockWsgiApp)
            assert vapp.app_for_testing.name_for_testing == 'app'
            assert vapp.app_for_testing.config_uri_for_testing == 'development.ini'
            assert vapp.environ_for_testing == {"HTTP_ACCEPT": "application/json", "REMOTE_USER": "TEST"}

            # The default is {"HTTP_ACCEPT": "application/json"}. If overridden, that must be included explicitly,
            # just in case the caller wants to omit it.
            some_environ = {"SOMETHING": "ELSE"}

            vapp = make_dev_vapp(environ=some_environ)
            assert vapp.app_for_testing.name_for_testing == 'app'
            assert vapp.environ_for_testing == {"REMOTE_USER": "TEST", "SOMETHING": "ELSE"}

            vapp = make_dev_vapp(environ={})
            assert vapp.app_for_testing.name_for_testing == 'app'
            assert vapp.environ_for_testing == {"REMOTE_USER": "TEST"}

            some_user = "some_user@cgap.hms.harvard.edu"

            vapp = make_dev_vapp(environ=some_environ, remote_user=some_user)
            assert vapp.app_for_testing.name_for_testing == 'app'
            assert vapp.environ_for_testing == {"REMOTE_USER": some_user, "SOMETHING": "ELSE"}

            vapp = make_dev_vapp(remote_user=some_user)
            assert vapp.app_for_testing.name_for_testing == 'app'
            assert vapp.environ_for_testing == {"HTTP_ACCEPT": "application/json", "REMOTE_USER": some_user}

            # It's possible also to get no remote user, by expressly passing None, since the default is TEST.
            # In that case, we expect no binding for REMOTE_USER in the environ:

            vapp = make_dev_vapp(remote_user=None)
            assert vapp.app_for_testing.name_for_testing == 'app'
            assert vapp.environ_for_testing == {"HTTP_ACCEPT": "application/json"}

            vapp = make_dev_vapp(remote_user=None, environ={})
            assert vapp.app_for_testing.name_for_testing == 'app'
            assert vapp.environ_for_testing == {}



from dcicutils.misc_utils import utc_now_str
from snovault.schema_validation import NO_DEFAULT
from .server_defaults_user import get_userid


def get_now():
    """ Wrapper for the server_default 'now' above so it is not called through SERVER_DEFAULTS in our code """
    return utc_now_str()


def add_last_modified(properties, **kwargs):
    """
        Uses the above two functions to add the last_modified information to the item
        May have no effect
        Allow someone to override the request userid (none in this case) by passing in a different uuid
        CONSIDER: `last_modified` (and `last_text_edited`) are not really 'server defaults' but rather system-managed fields.
    """

    userid = kwargs.get("userid", None)
    field_name_portion = kwargs.get("field_name_portion", "modified")

    last_field_name = "last_" + field_name_portion  # => last_modified
    by_field_name = field_name_portion + "_by"      # => modified_by
    date_field_name = "date_" + field_name_portion  # => date_modified

    try:
        last_modified = {
            by_field_name: get_userid(),
            date_field_name: get_now(),
        }
    except AttributeError:  # no request in scope ie: we are outside the core application.
        if userid:
            last_modified = {
                by_field_name: userid,
                date_field_name: get_now(),
            }
            properties[last_field_name] = last_modified
    else:
        # get_userid returns NO_DEFAULT if no userid
        if last_modified[by_field_name] != NO_DEFAULT:
            properties[last_field_name] = last_modified



from functools import lru_cache
import json
import sys
from typing import Optional
import uptime

from collections import OrderedDict
from dcicutils import lang_utils
from dcicutils.s3_utils import HealthPageKey
from dcicutils.env_utils import infer_foursight_url_from_env
from pyramid.decorator import reify
from pyramid.security import ALL_PERMISSIONS, Allow, Authenticated, Deny, Everyone
from . import Root, calculated_property, root, COLLECTIONS, STORAGE
from .appdefs import APP_VERSION_REGISTRY_KEY, ITEM_INDEX_ORDER
from .schema_formats import is_accession
from .util import SettingsKey


def includeme(config):
    config.include(health_check)
    config.include(item_counts)
    config.include(type_metadata)
    config.include(submissions_page)
    config.scan(__name__)


def item_counts(config):
    config.add_route(
        'item-counts',
        '/counts'
    )

    def counts_view(request):
        response = request.response
        response.content_type = 'application/json; charset=utf-8'

        db_total = 0
        es_total = 0
        # find db and es counts for each item type
        db_es_compare = OrderedDict()
        for item_type in request.registry[COLLECTIONS].by_item_type:
            # use the write (DB) storage with only the specific item_type
            # need to count items with props in ES differently
            db_count = request.registry[STORAGE].write.__len__(item_type)
            es_count = request.registry[STORAGE].read.__len__(item_type)
            db_total += db_count
            es_total += es_count
            warn_str = build_warn_string(db_count, es_count)
            item_name = request.registry[COLLECTIONS][item_type].type_info.name
            db_es_compare[item_name] = ("DB: %s   ES: %s %s" % (str(db_count), str(es_count), warn_str))
        warn_str = build_warn_string(db_total, es_total)
        db_es_total = ("DB: %s   ES: %s %s" %
                       (str(db_total), str(es_total), warn_str))
        response_dict = {
            'title': 'Item Counts',
            'db_es_total': db_es_total,
            'db_es_compare': db_es_compare
        }
        return response_dict

    config.add_view(counts_view, route_name='item-counts')


def type_metadata(config):
    """ This needs to be re-written in the upstream repo """

    config.add_route(
        'type-metadata',
        '/type-metadata'
    )

    def type_metadata_view(request):

        return {
            'index_order': ITEM_INDEX_ORDER
        }

    config.add_view(type_metadata_view, route_name='type-metadata')



def uptime_info():
    try:
        return lang_utils.relative_time_string(uptime.uptime())
    except Exception:
        return "unavailable"


def health_check(config):
    """
    Emulate a lite form of Alex's static page routing
    """
    config.add_route(
        'health-check',
        '/health'
    )

    def health_page_view(request):

        class ExtendedHealthPageKey(HealthPageKey):
            # This class can contain new entries in HealthPageKey that are waiting to move to dcicutils
            PYTHON_VERSION = "python_version"
            pass

        h = ExtendedHealthPageKey

        s = SettingsKey

        response = request.response
        response.content_type = 'application/json; charset=utf-8'
        settings = request.registry.settings

        env_name = settings.get('env.name')
        foursight_url = infer_foursight_url_from_env(request=request, envname=env_name)

        response_dict = {

            "@type": ["Health", "Portal"],
            "@context": "/health",
            "@id": "/health",
            "content": None,

            h.APPLICATION_BUCKET_PREFIX: settings.get(s.APPLICATION_BUCKET_PREFIX),
            h.BEANSTALK_APP_VERSION: settings.get(s.EB_APP_VERSION),
            h.BEANSTALK_ENV: env_name,
            h.BLOB_BUCKET: settings.get(s.BLOB_BUCKET),
            h.DATABASE: settings.get(s.SQLALCHEMY_URL).split('@')[1],  # don't show user /password
            h.DISPLAY_TITLE: "ENCODED Portal Status and Foursight Monitoring",
            h.ELASTICSEARCH: settings.get(s.ELASTICSEARCH_SERVER),
            h.FILE_UPLOAD_BUCKET: settings.get(s.FILE_UPLOAD_BUCKET),
            h.FOURSIGHT: foursight_url,
            h.FOURSIGHT_BUCKET_PREFIX: settings.get(s.FOURSIGHT_BUCKET_PREFIX),
            h.IDENTITY: settings.get(s.IDENTITY),
            h.INDEXER: settings.get(s.INDEXER),
            h.INDEX_SERVER: settings.get(s.INDEX_SERVER),
            h.LOAD_DATA: settings.get(s.LOAD_TEST_DATA),
            h.METADATA_BUNDLES_BUCKET: settings.get(s.METADATA_BUNDLES_BUCKET),
            h.NAMESPACE: settings.get(s.INDEXER_NAMESPACE),
            h.PROCESSED_FILE_BUCKET: settings.get(s.FILE_WFOUT_BUCKET),
            h.PROJECT_VERSION: settings.get(s.ENCODED_VERSION),
            h.PYTHON_VERSION: f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            h.S3_ENCRYPT_KEY_ID: settings.get(s.S3_ENCRYPT_KEY_ID),
            h.SNOVAULT_VERSION: settings.get(s.SNOVAULT_VERSION),
            h.SYSTEM_BUCKET: settings.get(s.SYSTEM_BUCKET),
            h.TIBANNA_CWLS_BUCKET: settings.get(s.TIBANNA_CWLS_BUCKET),
            h.TIBANNA_OUTPUT_BUCKET: settings.get(s.TIBANNA_OUTPUT_BUCKET),
            h.UPTIME: uptime_info(),
            h.UTILS_VERSION: settings.get(s.UTILS_VERSION),
            h.GIT: _get_gitinfo()

        }

        return response_dict

    config.add_view(health_page_view, route_name='health-check')


@lru_cache(maxsize=1)
def _get_gitinfo() -> Optional[dict]:
    try:
        with open("gitinfo.json") as f:
            return json.load(f)
    except Exception:
        return None


def build_warn_string(db_count, es_count):
    if db_count > es_count:
        warn_str = '  < DB has %s more items >' % (str(db_count - es_count))
    elif db_count < es_count:
        warn_str = '  < ES has %s more items >' % (str(es_count - db_count))
    else:
        warn_str = ''
    return warn_str


def submissions_page(config):
    """
    Emulate a lite form of Alex's static page routing
    """
    config.add_route(
        'submissions-page',
        '/submissions'
    )

    def submissions_page_view(request):
        response = request.response
        response.content_type = 'application/json; charset=utf-8'

        response_dict = {
            "title": "Submissions",
            "notification": "success",
            "@type": ["Submissions", "Portal"],
            "@context": "/submissions",
            "@id": "/submissions",
            "content": None
        }

        return response_dict

    config.add_view(submissions_page_view, route_name='submissions-page')


def acl_from_settings(settings):
    # XXX Unsure if any of the demo instance still need this
    acl = []
    for k, v in settings.items():
        if k.startswith('allow.'):
            action = Allow
            permission = k[len('allow.'):]
            principals = v.split()
        elif k.startswith('deny.'):
            action = Deny
            permission = k[len('deny.'):]
            principals = v.split()
        else:
            continue
        if permission == 'ALL_PERMISSIONS':
            permission = ALL_PERMISSIONS
        for principal in principals:
            if principal == 'Authenticated':
                principal = Authenticated
            elif principal == 'Everyone':
                principal = Everyone
            acl.append((action, principal, permission))
    return acl


@root
class SnovaultRoot(Root):
    properties = {
        'title': 'Home',
        'portal_title': 'Snovault',
    }

    @reify
    def __acl__(self):
        acl = acl_from_settings(self.registry.settings) + [
            (Allow, Everyone, ['list', 'search']),
            (Allow, 'group.admin', ALL_PERMISSIONS),
            (Allow, 'remoteuser.EMBED', 'restricted_fields'),
        ] + [
            (Allow, 'remoteuser.INDEXER', ['view', 'view_raw', 'list', 'index']),
            (Allow, 'remoteuser.EMBED', ['view', 'view_raw', 'expand']),
            (Allow, Everyone, ['visible_for_edit'])
        ]
        return acl

    def get(self, name, default=None):
        resource = super().get(name, None)
        if resource is not None:
            return resource
        resource = self.connection.get_by_unique_key('page:location', name)
        if resource is not None:
            return resource
        if is_accession(name):
            resource = self.connection.get_by_unique_key('accession', name)
            if resource is not None:
                return resource
        if ':' in name:
            resource = self.connection.get_by_unique_key('alias', name)
            if resource is not None:
                return resource
        return default

    def get_by_uuid(self, uuid, default=None):
        return self.connection.get_by_uuid(uuid, default)

    def jsonld_context(self):
        """Inherits from '@context' calculated property of Resource in snovault/resources.py"""
        return '/home'

    def jsonld_type(self):
        """Inherits from '@type' calculated property of Root in snovault/resources.py"""
        return ['HomePage', 'StaticPage'] + super().jsonld_type()

    @calculated_property(schema={
        "title": "Static Page Content",
        "type": "array"
    })
    def content(self, request):
        """Returns -object- with pre-named sections"""
        return []
        # sections_to_get = ['home.introduction']
        # user = request._auth0_authenticated if hasattr(request, '_auth0_authenticated') else True
        # return_list = []
        # for section_name in sections_to_get:
        #     try:  # Can be caused by 404 / Not Found during indexing
        #         res = request.embed('/static-sections', section_name, '@@embedded', as_user=user)
        #         return_list.append(res)
        #     except KeyError:
        #         pass
        # return return_list

    @calculated_property(schema={
        "title": "Application version",
        "type": "string",
    })
    def app_version(self, registry):
        return registry.settings[APP_VERSION_REGISTRY_KEY]



# Considation of ACL related definitions.

from pyramid.security import Allow, Deny, Everyone
from typing import Any, List, Tuple, Union

Acl = List[Tuple[Any, Any, Union[str, List[str]]]]

ONLY_ADMIN_VIEW_ACL: Acl = [
    (Allow, 'group.admin', ['view', 'edit']),
    (Allow, 'group.read-only-admin', ['view']),
    (Allow, 'remoteuser.INDEXER', ['view']),
    (Allow, 'remoteuser.EMBED', ['view']),
    (Deny, Everyone, ['view', 'edit'])
]

PUBLIC_ACL: Acl = [
    (Allow, Everyone, ['view']),
] + ONLY_ADMIN_VIEW_ACL

DELETED_ACL: Acl = [
    (Deny, Everyone, 'visible_for_edit')
] + ONLY_ADMIN_VIEW_ACL

# Originally from user.py:

ONLY_ADMIN_VIEW_USER_DETAILS_ACL = [
    (Allow, 'group.admin', ['view', 'view_details', 'edit']),
    (Allow, 'remoteuser.INDEXER', ['view']),
    (Allow, 'remoteuser.EMBED', ['view']),
    (Deny, Everyone, ['view', 'view_details', 'edit']),
]

ONLY_OWNER_VIEW_PROFILE_ACL = [
    (Allow, 'role.owner', 'view'),
    # (Allow, 'role.owner', 'edit'),
    # (Allow, 'role.owner', 'view_details'),
] + ONLY_ADMIN_VIEW_USER_DETAILS_ACL

DELETED_USER_ACL = [
    (Deny, Everyone, 'visible_for_edit')
] + ONLY_ADMIN_VIEW_USER_DETAILS_ACL




import pytest
import re
from snovault.ingestion.ingestion_message import IngestionMessage
from snovault.ingestion.ingestion_listener_base import IngestionListenerBase
from snovault.ingestion.ingestion_message_handler_decorator import (
    call_ingestion_message_handler,
    ingestion_message_handler,
    clear_ingestion_message_handlers_for_testing,
)


class IngestionListener(IngestionListenerBase):
    pass  # dummy


SOME_UUID = "some-uuid-xyzzy"
INGESTION_LISTENER = IngestionListener()
INGESTION_TYPE_VCF = "vcf"
INGESTION_TYPE_NOVCF = "novcf"
INGESTION_TYPE_OTHER = "other"


def isolate_ingestion_message_handler_decorator_test(f):
    def wrapper():
        clear_ingestion_message_handlers_for_testing()
        f()
        clear_ingestion_message_handlers_for_testing()
    return wrapper


def create_raw_message(ingestion_type: str) -> dict:
    return {"Body": f"{{\"uuid\":\"{SOME_UUID}\", \"ingestion_type\":\"{ingestion_type}\"}}"}


@isolate_ingestion_message_handler_decorator_test
def test_error_decorator_arguments():

    with pytest.raises(Exception):
        @ingestion_message_handler(123)  # wrong decorator arg type
        def bad_a(message, listener):
            pass

    with pytest.raises(Exception):
        @ingestion_message_handler("vcf", 123)  # too many decorator args
        def bad_b(message, listener):
            pass

    with pytest.raises(Exception):
        @ingestion_message_handler(xyzzy="vcf")  # unknown named decorator kwarg
        def bad_c(message, listener):
            pass

    with pytest.raises(Exception):
        @ingestion_message_handler("vcf", ingestion_type="vcf")  # too many decorator args
        def bad_d(message, listener):
            pass


@isolate_ingestion_message_handler_decorator_test
def test_error_decorated_function_signature():

    with pytest.raises(Exception):
        @ingestion_message_handler
        def bad_a():  # not enough args
            pass

    with pytest.raises(Exception):
        @ingestion_message_handler
        def bad_b(message):  # not enough args
            pass

    with pytest.raises(Exception):
        @ingestion_message_handler
        def bad_c(message, listener, extraneous_arg):  # too many args
            pass

    with pytest.raises(Exception):
        @ingestion_message_handler
        def bad_d(message: IngestionMessage, listener: str):  # wrong type arg
            pass

    with pytest.raises(Exception):
        @ingestion_message_handler
        def bad_e(message: str, listener):  # wrong type arg
            pass

    with pytest.raises(Exception):
        @ingestion_message_handler
        def bad_f(message: str, listener: str):  # wrong type args
            pass


@isolate_ingestion_message_handler_decorator_test
def test_error_duplicate_default_handlers_one():

    with pytest.raises(Exception):

        @ingestion_message_handler
        def a(message, listener):
            pass

        @ingestion_message_handler  # same as above (i.e. default)
        def duplicate_a(message, listener):
            pass


@isolate_ingestion_message_handler_decorator_test
def test_error_duplicate_default_handlers_two():

    with pytest.raises(Exception):

        @ingestion_message_handler(ingestion_type="default")
        def a(message, listener):
            pass

        @ingestion_message_handler  # same as above (i.e. default)
        def duplicate_a(message, listener):
            pass


@isolate_ingestion_message_handler_decorator_test
def test_error_duplicate_typed_handlers():

    with pytest.raises(Exception):

        @ingestion_message_handler("some-message-type")
        def a(message, listener):
            pass

        @ingestion_message_handler("some-message-type")  # same as above
        def duplicate_a(message, listener):
            pass


@isolate_ingestion_message_handler_decorator_test
def test_error_undefined_handler():

    with pytest.raises(Exception):

        @ingestion_message_handler("some-message-type")
        def a(message, listener):
            pass

        @ingestion_message_handler("some-other-message-type")
        def duplicate_a(message, listener):
            pass

        ingestion_message = create_raw_message(ingestion_type="some-third-message-type")
        # This should throw exception because no relevant handler found.
        call_ingestion_message_handler(ingestion_message, INGESTION_LISTENER)

    exception_message = (f".*No.*ingestion.*message.*handler.*defined.*"
                         f" ->.*{{'uuid': 'some-uuid-xyzzy', 'ingestion_type': 'some-third-message-type'}}.*")
    with pytest.raises(Exception, match=re.compile(exception_message)):
        call_ingestion_message_handler(ingestion_message, INGESTION_LISTENER)


@isolate_ingestion_message_handler_decorator_test
def test_error_invalid_call_arguments():

    @ingestion_message_handler("some-message-type")
    def a(message, listener):
        pass

    ingestion_message = create_raw_message(ingestion_type="some-message-type")
    with pytest.raises(Exception):
        call_ingestion_message_handler(ingestion_message, "wrong-type-should-be-IngestionListenerBase")

    ingestion_message = create_raw_message(ingestion_type="some-message-type")
    with pytest.raises(Exception):
        call_ingestion_message_handler(ingestion_message, INGESTION_LISTENER, "extra-arg")

    ingestion_message = create_raw_message(ingestion_type="some-message-type")
    with pytest.raises(Exception):
        call_ingestion_message_handler(ingestion_message)  # missing IngestionListenerBase arg

    ingestion_message = create_raw_message(ingestion_type="some-message-type")
    with pytest.raises(Exception):
        call_ingestion_message_handler()  # missing args

    with pytest.raises(Exception):
        a(ingestion_message, INGESTION_LISTENER)  # wrong first arg type (raw dict rather than IngestionMessage)


@isolate_ingestion_message_handler_decorator_test
def test_one():

    handler_calls = None

    @ingestion_message_handler
    def a(message: IngestionMessage, listener: IngestionListener):
        this_function_name = "a"
        result = f"{this_function_name}/{message.type}"
        handler_calls.add(result)
        assert not message.is_type(INGESTION_TYPE_VCF) and not message.is_type(INGESTION_TYPE_NOVCF)
        assert message.uuid == SOME_UUID
        assert listener is INGESTION_LISTENER
        return result

    @ingestion_message_handler(ingestion_type=INGESTION_TYPE_VCF)
    def b(message: IngestionMessage, listener: IngestionListener) -> str:
        this_function_name = "b"
        result = f"{this_function_name}/{message.type}"
        handler_calls.add(result)
        assert message.is_type(INGESTION_TYPE_VCF)
        assert message.uuid == SOME_UUID
        assert listener is INGESTION_LISTENER
        return result

    @ingestion_message_handler(INGESTION_TYPE_NOVCF)
    def c(message, listener):
        this_function_name = "c"
        result = f"{this_function_name}/{message.type}"
        handler_calls.add(result)
        assert message.is_type(INGESTION_TYPE_NOVCF)
        assert message.uuid == SOME_UUID
        assert listener is INGESTION_LISTENER
        return result

    handler_calls = set()
    ingestion_message = create_raw_message(ingestion_type=INGESTION_TYPE_OTHER)
    handler_result = call_ingestion_message_handler(ingestion_message, INGESTION_LISTENER)
    assert handler_result == f"a/{IngestionMessage(ingestion_message).type}"
    assert handler_calls == {handler_result}

    handler_calls = set()
    ingestion_message = create_raw_message(ingestion_type=INGESTION_TYPE_VCF)
    handler_result = call_ingestion_message_handler(ingestion_message, INGESTION_LISTENER)
    assert handler_result == f"b/{IngestionMessage(ingestion_message).type}"
    assert handler_calls == {handler_result}

    handler_calls = set()
    ingestion_message = create_raw_message(ingestion_type=INGESTION_TYPE_NOVCF)
    handler_result = call_ingestion_message_handler(ingestion_message, INGESTION_LISTENER)
    assert handler_result == f"c/{IngestionMessage(ingestion_message).type}"
    assert handler_calls == {handler_result}



# Module containing the definition of the @ingestion_message_handler decorator used
# to globally register ingestion message handler functions for specific ingestion
# message types, or a default message handler to handle any message types for which
# no specific handler was registered. Only a single handler may be registered for
# a specified message type, and only a single default handler may be registered.
# Also defined here is a function (call_ingestion_message_handler) to call the
# appropriate registered handler for a given message.

import inspect
from typing import Union
from dcicutils.misc_utils import ignored, PRINT
from .ingestion_listener_base import IngestionListenerBase
from .ingestion_message import IngestionMessage


# Dictionary (by ingestion type) of globally registered ingestion message handlers.
_ingestion_message_handlers = {}


def ingestion_message_handler(f=None, *decorator_args, **decorator_kwargs):
    """
    Decorator to globally register ingestion message handlers, to be used for example like this:

      @ingestion_message_handler
      def your_ingester_message_handler(message: IngestionMessage, listener: IngestionListener):
          # Handle your message here; return whatever you like;
          # it will be returned in turn by call_ingestion_message_handler.

    Although any function may be annotated with this decorator, at this time and for our purposes
    it is expected to have a signature as show in the example above; this IS enforced to some extent.

    In addition, you can pass an ingestion_type argument to the decorator to LIMIT the call of the
    decorated handler function to messages with an ingestion type which matches the specified string value.
    For example, to define a message handler to be called ONLY for message types which are "vcf":

      @ingestion_message_handler(ingestion_type="vcf")
      def your_ingester_message_handler(message: IngestionMessage, listener: IngestionListener):
          # Handle your message here; return whatever you like;
          # it will be returned in turn by call_ingestion_message_handler.

    Note that ingestion type names are (space-trimmed and) treated as case-insenstive.

    If the ingestion_type is not specified in the decorator, then the registered handler is said
    to be the DEFAULT handler and will handle any message type not covered by any other registered
    handler. There MUST be exactly ONE handler registered able to handle any expected message type,
    otherwise an exception will be thrown (either at handler registration time, i.e. startup; or
    at runtime, i.e. if an incoming message is found not to have an associated handler and there is
    no default handler; this latter bit is handled by the call_ingestion_message_handler function below).
    """
    ignored(decorator_args)
    has_decorator_args = True if not callable(f) or f.__name__ == "<lambda>" else False
    ingestion_type = None

    # Sanity check any decorator arguments; currently just the optional ingestion_type.
    if has_decorator_args:
        if f is not None:
            decorator_args = (f, *decorator_args)
        if len(decorator_args) + len(decorator_kwargs) > 1:
            raise ValueError(f"Invalid @ingestion_message_handler decorator usage (takes at most one argument).")
        if len(decorator_args) == 1:
            ingestion_type = decorator_args[0]
        else:
            ingestion_type = decorator_kwargs.get("ingestion_type", decorator_kwargs.get("type"))
        if not (ingestion_type is None or isinstance(ingestion_type, str)):
            raise ValueError(f"Invalid @ingestion_message_handler decorator usage (argument must be ingestion type string).")
        ingestion_type = ingestion_type.strip().lower()
    # If ingestion_type is not specified or is "default" this we are registering a default handler.
    if ingestion_type == "default":
        ingestion_type = None

    def ingestion_message_handler_wrapper(wrapped_function):

        if ingestion_type in _ingestion_message_handlers:
            raise ValueError(f"Ingestion message handler already defined for "
                             f"ingestion message type: {ingestion_type if ingestion_type else '<default>'}")

        # Sanity check the signature of the decorated ingestion message handler function.
        # It should contain two arguments with either no type annotations or if present
        # then they should be for IngestionMessage and IngestionListenerBase, respectively.
        # Return value annotation is not checked.
        wrapped_function_signature = inspect.signature(wrapped_function)
        if len(wrapped_function_signature.parameters) < 2:
            raise ValueError(f"Too few arguments (need two) "
                             f"for ingestion message handler function: {wrapped_function.__name__}")
        if len(wrapped_function_signature.parameters) > 2:
            raise ValueError(f"Too many arguments (need two) "
                             f"for ingestion message handler function: {wrapped_function.__name__}")
        parameters = iter(wrapped_function_signature.parameters.items())
        first_parameter = next(parameters)
        if first_parameter and len(first_parameter) >= 2:
            first_parameter_annotation = first_parameter[1].annotation
            if not first_parameter_annotation or (first_parameter_annotation.__name__ != "_empty" and
                                                  not issubclass(first_parameter_annotation, IngestionMessage)):
                raise ValueError(f"Wrong first argument type (need unspecified or IngestionMessage) "
                                 f"for ingestion message handler function: {wrapped_function.__name__}")
        second_parameter = next(parameters)
        if second_parameter and len(second_parameter) >= 2:
            second_parameter_annotation = second_parameter[1].annotation
            if not second_parameter_annotation or (second_parameter_annotation.__name__ != "_empty" and
                                                   not issubclass(second_parameter_annotation, IngestionListenerBase)):
                raise ValueError(f"Wrong second argument type (need unspecified or IngestionListenerBase) "
                                 f"for ingestion message handler function: {wrapped_function.__name__}")
        PRINT(f"Registering ingestion message handler: "
              f"{wrapped_function.__name__} (type: {ingestion_type if ingestion_type else '<default>'})")

        def ingestion_message_handler_function(*args, **kwargs):
            """
            This is the function called on each actual ingestion message handler call.
            """
            ignored(kwargs)
            # Check for two arguments of type IngestionMessage and IngestionListenerBase, respectively.
            if len(args) != 2:
                raise ValueError(f"Wrong number of arguments ({len(args)} passed to "
                                 f"ingestion message handler (expecting two): {wrapped_function.__name__}")
            message = args[0]
            listener = args[1]
            if not isinstance(message, IngestionMessage):
                raise ValueError(f"First argument passed to ingestion message handler is "
                                 f"not of type IngestionMessage: {wrapped_function.__name__}")
            if not isinstance(listener, IngestionListenerBase):
                raise ValueError(f"Second argument passed to ingestion message handler is "
                                 f"not of type IngestionListenerBase: {wrapped_function.__name__}")
            # Ensure we should call this handler based on any ingestion_type specified in the decorator.
            # Given the current implementation and intended usage (i.e. handlers be specifically associated
            # with a given message type, and calling via call_ingestion_message_handler) this should check
            # should be unnecessary, though extra check will not hurt; it would only come up if calling a
            # registered message handler directly (i.e. not via call_ingestion_message_handler).
            PRINT(f"Checking message ({message.uuid}) type ({message.type}) for handler: {wrapped_function.__name__}")
            if ingestion_type:
                # Here the decorator specified a NON-default ingestion type for this handler;
                # check and only call this handler (the wrapped function) if the handler
                # ingestion type matches the ingestion message type.
                if not message.is_type(ingestion_type):
                    # Since the ingestion_type specified for the handler decorator does NOT match
                    # the type of the message, then this message is NOT intended to be processed by
                    # this handler, it will NOT be called. Again, as mentioned above, this should
                    # NOT come up if the handler is called via call_ingestion_message_handler.
                    PRINT(f"Message ({message.uuid}) type ({message.type}) "
                          f"NOT intended for handler: {wrapped_function.__name__}")
                    return False
            # Here this handler decorator either had no ingestion_type specifier, or it does
            # and it matches the ingestion message type, indicating this message IS intended
            # to be processed by this handler; we will call it here, returning its value.
            PRINT(f"Calling message ({message.uuid}) type ({message.type}) "
                  f"handler: {wrapped_function.__name__}")
            handler_result = wrapped_function(message, listener)
            PRINT(f"Called message ({message.uuid}) type ({message.type}) "
                  f"handler: {wrapped_function.__name__} -> {handler_result}")
            return handler_result

        # Register this handler for the ingestion type in our global dictionary;
        # already checked above if a handler is already registered for this type.
        _ingestion_message_handlers[ingestion_type] = ingestion_message_handler_function

        return ingestion_message_handler_function

    return ingestion_message_handler_wrapper(f) if not has_decorator_args else ingestion_message_handler_wrapper


def call_ingestion_message_handler(message: Union[IngestionMessage, dict], listener) -> bool:
    """
    Calls the ingestion message handler function globally registered via the
    @ingestion_message_handler decorator which corresponding to the TYPE of the given
    IngestionMessage, passing it the given IngestionMessage and IngestionListenerBase
    as arguments; returns the value returned by the message handler.

    If a message handler has NOT been registered for the given message type AND of NO default
    message handler has been registered, then throws and exception. I.e. a specific message handler
    MUST be defined for each expected message type OR a DEFAULT message handler must be defined
    to handle messages with types which does NOT correspond to any specifically registered handlers.
    """
    if not isinstance(message, IngestionMessage):
        # For convenience, allow passing a message which is NOT of type IngestionMessage, which we
        # will ASSUME in this case is a RAW (dict) message from which we create an IngestionMessage.
        message = IngestionMessage(message)
    # Get the handler for this message type, or the default handler of none specifically found.
    handler = _ingestion_message_handlers.get(message.type, _ingestion_message_handlers.get(None))
    if handler:
        return handler(message, listener)
    else:
        # If NO message handler is registered for the given message type AND if there
        # is NO default message handler registered then we regard this as a (runtime) error.
        raise RuntimeError(f"No ingestion message handler defined for ingestion message type: {message.type}"
                           f" -> Message: {message.body}")


def clear_ingestion_message_handlers_for_testing():
    """
    Clears all globally registered ingestion message handlers.
    This is for TESTING purposes ONLY!
    """
    global _ingestion_message_handlers
    _ingestion_message_handlers = {}



import time
import json
import socket
import boto3
import structlog
import datetime


log = structlog.getLogger(__name__)


class IngestionQueueManager:
    """
    Similar to QueueManager in snovault in that in manages SQS queues, but that code is not generic
    enough to use here, so it is "duplicated" so to speak here. At a later time the functionality of this
    class and QueueManager should be refactored into a "helper" class, but for now this is sufficient
    and is tested independently here.

    We will use a single queue to keep track of File uuids to be indexed. This used to manage only VCFs
    but now the Ingestion functionality is generic and can be extended to arbitrary processing on
    any type.
    """
    QUEUE_NAME_EXTENSION = '-ingestion-queue'  # XXX: breaking change, matches 4dn-cloud-infra resources

    def __init__(self, registry, override_name=None):
        """ Does initial setup for interacting with SQS """
        self.batch_size = 1  # NOTE: this value is important because we don't want to block other jobs
        self.env_name = registry.settings.get('env.name', None)
        if not self.env_name:  # replace with something usable
            backup = socket.gethostname()[:80].replace('.', '-')
            self.env_name = backup if backup else 'cgap-backup'
        kwargs = {
            'region_name': 'us-east-1'
        }
        self.client = boto3.client('sqs', **kwargs)
        self.queue_name = override_name or (self.env_name + self.QUEUE_NAME_EXTENSION)
        self.queue_attrs = {
            self.queue_name: {
                'DelaySeconds': '1',  # messages initially invisible for 1 sec
                'VisibilityTimeout': '10800',  # 3 hours
                'MessageRetentionPeriod': '604800',  # 7 days, in seconds
                'ReceiveMessageWaitTimeSeconds': '5',  # 5 seconds of long polling
            }
        }
        self.queue_url = self._initialize()

    def _initialize(self):
        """ Initializes the actual queue - helper method for init """
        try:
            response = self.client.create_queue(
                QueueName=self.queue_name,
                Attributes=self.queue_attrs[self.queue_name]
            )
            queue_url = response['QueueUrl']
        except self.client.exceptions.QueueNameExists:
            queue_url = self._get_queue_url(self.queue_name)
        except Exception as e:
            log.error('Error while attempting to create queue: %s' % e)
            queue_url = self._get_queue_url(self.queue_name)
        return queue_url

    def _get_queue_url(self, queue_name):
        """
        Simple function that returns url of associated queue name
        """
        try:
            response = self.client.get_queue_url(
                QueueName=queue_name
            )
        except Exception as e:
            log.error('Cannot resolve queue_url: %s' % e)
            response = {}
        return response.get('QueueUrl', None)

    def _chunk_messages(self, msgs):
        """ Chunks messages into self.send_batch_size batches (for efficiency).

        :param msgs: list of messages to be chunked
        """
        for i in range(0, len(msgs), self.batch_size):
            yield msgs[i:i + self.batch_size]

    def _send_messages(self, msgs, retries=3):
        """ Sends msgs to the ingestion queue (with retries for failed messages).

        :param msgs: to be sent
        :param retries: number of times to resend failed messages, decremented on recursion
        :return: list of any failed messages
        """
        failed = []
        for msg_batch in self._chunk_messages(msgs):
            log.info('Trying to chunk messages: %s' % msgs)
            entries = []
            for msg in msg_batch:
                entries.append({
                    'Id': str(int(time.time() * 1000000)),
                    'MessageBody': json.dumps(msg)
                })
            response = self.client.send_message_batch(
                QueueUrl=self.queue_url,
                Entries=entries
            )
            failed_messages = response.get('Failed', [])

            # attempt resend of failed messages
            if failed_messages and retries > 0:
                msgs_to_retry = []
                for failed_message in failed_messages:
                    fail_id = failed_message.get('Id')
                    msgs_to_retry.extend([json.loads(ent['MessageBody']) for ent in entries if ent['Id'] == fail_id])
                    if msgs_to_retry:
                        failed_messages = self._send_messages(msgs_to_retry, retries=retries - 1)
            failed.extend(failed_messages)
        return failed

    def delete_messages(self, messages):
        """
        Called after a message has been successfully received and processed.
        Removes message from the queue.
        Input should be the messages directly from receive messages. At the
        very least, needs a list of messages with 'Id' and 'ReceiptHandle' as this
        metadata is necessary to identify the message in SQS internals.

        NOTE: deletion does NOT have a retry mechanism

        :param messages: messages to be deleted
        :returns: a list with any failed messages
        """
        failed = []
        for batch in self._chunk_messages(messages):
            # need to change message format, since deleting takes slightly
            # different fields what's return from receiving
            for i in range(len(batch)):
                to_delete = {
                    'Id': batch[i]['MessageId'],
                    'ReceiptHandle': batch[i]['ReceiptHandle']
                }
                batch[i] = to_delete
            response = self.client.delete_message_batch(
                QueueUrl=self.queue_url,
                Entries=batch
            )
            failed.extend(response.get('Failed', []))
        return failed

    def add_uuids(self, uuids, ingestion_type='vcf'):
        """ Takes a list of string uuids and adds them to the ingestion queue.
            If ingestion_type is not specified, it defaults to 'vcf'.

            :precondition: uuids are all of type FileProcessed
            :param uuids: uuids to be added to the queue.
            :param ingestion_type: the ingestion type of the uuids (default 'vcf' for legacy reasons)
            :returns: 2-tuple: uuids queued, failed messages (if any)
        """
        curr_time = datetime.datetime.utcnow().isoformat()
        msgs = []
        for uuid in uuids:
            current_msg = {
                'ingestion_type': ingestion_type,
                'uuid': uuid,
                'timestamp': curr_time
            }
            msgs.append(current_msg)
        failed = self._send_messages(msgs)
        return uuids, failed

    def get_counts(self):
        """ Returns number counts of waiting/inflight messages
            * Makes a boto3 API Call to do so *

            :returns: 2 tuple of waiting, inflight messages
        """
        response = self.client.get_queue_attributes(
            QueueUrl=self.queue_url,
            AttributeNames=[
                'ApproximateNumberOfMessages',
                'ApproximateNumberOfMessagesNotVisible'
            ]
        )
        formatted = {
            'waiting': response.get('Attributes', {}).get('ApproximateNumberOfMessages'),
            'inflight': response.get('Attributes', {}).get('ApproximateNumberOfMessagesNotVisible')
        }
        return formatted['waiting'], formatted['inflight']

    def receive_messages(self, batch_size=None):
        """ Returns an array of messages, if any that are waiting

            :param batch_size: an integer number of messages
            :returns: messages received or [] if no messages were ready to be received
        """
        response = self.client.receive_message(
            QueueUrl=self.queue_url,
            MaxNumberOfMessages=self.batch_size if batch_size is None else batch_size
        )
        return response.get('Messages', [])

    def clear_queue(self):
        """ Clears the queue by receiving all messages. BE CAREFUL as this has potential to
            infinite loop under certain conditions. This risk is preferred to using 'purge', which
            has a long timeout. The guarantees this functions provides are minimal at best - it should
            really only be used in testing.
        """
        while True:
            messages = self.receive_messages()
            self.delete_messages(messages)
            if len(messages) == 0:
                break
